/**
 * MySDL - A multimedia framework based on SDL and OpenGL
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://www.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "loaders.h"

namespace mysdl
{
	namespace audio
	{
		const ALsizei DEF_BUFFER_SIZE=65536;//*2;//*4;//4*4096*16*2;

		std::string getOggErrorString(int error)
		{
			switch(error)
			{
				case OV_FALSE:
					return(std::string("Not true, or no data available."));

				case OV_HOLE:
					return(std::string("Vorbisfile encoutered missing or corrupt data in the bitstream."));

				case OV_EREAD:
					return(std::string("Read error while fetching compressed data for decode."));

				case OV_EFAULT:
					return(std::string("Internal inconsistency in decode state."));

				case OV_EIMPL:
					return(std::string("Feature not implemented."));

				case OV_EINVAL:
					return(std::string("Either an invalid argument, or incompletely initialized argument passed to libvorbisfile call."));

				case OV_ENOTVORBIS:
					return(std::string("The given file/data was not recognized as Ogg Vorbis data."));

				case OV_EBADHEADER:
					return(std::string("The file/data is apparently an Ogg Vorbis stream, but contains a corrupted or undecipherable header."));

				case OV_EVERSION:
					return(std::string("The bitstream format revision of the given stream is not supported."));

				case OV_EBADLINK:
					return(std::string("The given link exists in the Vorbis data stream, but is not decipherable due to garbacge or corruption."));

				case OV_ENOSEEK:
					return(std::string("The given stream is not seekable."));

				default:
					return(std::string("(Unknown Ogg error)"));
			};
		}

		FileLoader::FileLoader(FILE *file,const std::string filename)
		{
//			//debug::addMessage("FileLoader::FileLoader()");

			file=0;
			format=0;
			size=0;
			frequency=0;
			bufferSize=0;
			buffer=0;

			if(file)
			{
				this->file=file;
			}
			else
			{
				if(!(this->file=fopen(filename.c_str(),"rb")))
				{
					throw std::string("Could not open file \"")+filename+std::string("\".");
				}
			}

			size=0;
			format=0;
			size=0;
			frequency=0;

			bufferSize=0;
			buffer=0;

//			//debug::removeMessage();
		}

		FileLoader::~FileLoader()
		{
//			//debug::addMessage("FileLoader::~FileLoader()");

			freeBuffer();

			if(file)
			{
				fclose(file);
				file=0;
			}

//			//debug::removeMessage();
		}

		void FileLoader::freeBuffer()
		{
//			//debug::addMessage("FileLoader::freeBuffer()");

			if(buffer)
			{
				delete[] buffer;
				buffer=0;
				bufferSize=0;
			}

//			//debug::removeMessage();
		}

		ALsizei FileLoader::getFormat()
		{
//mysdl::debug::critical(dbgAudio,"getFormat");
			return(format);
		}

		ALsizei FileLoader::getSize()
		{
			return(size);
		}

		ALsizei FileLoader::getFrequency()
		{
//mysdl::debug::critical(dbgAudio,"getFrequency");
			return(frequency);
		}

		OggFileLoader::OggFileLoader(FILE *file,const std::string filename) : FileLoader(file,filename),oggStream()
		{
//			//debug::addMessage("OggFileLoader::OggFileLoader()");

			vorbisInfo=0;

			int res=ov_open(this->file,&oggStream,0,0);

			if(res<0)
			{
				throw std::string("Could not read Ogg stream:")+getOggErrorString(res);
			}

			vorbisInfo=ov_info(&oggStream,-1);

			format=(vorbisInfo->channels==1 ? AL_FORMAT_MONO16 : AL_FORMAT_STEREO16);
			frequency=vorbisInfo->rate;
			size=(vorbisInfo->channels*2)*ov_pcm_total(&oggStream,-1);

//			//debug::removeMessage();
		}

		OggFileLoader::~OggFileLoader()
		{
//			//debug::addMessage("OggFileLoader::~OggFileLoader()");

			ov_clear(&oggStream);
			file=0;

//			//debug::removeMessage();
		}

		void OggFileLoader::rewind()
		{
//			//debug::addMessage("OggFileLoader::rewind()");
//			int res=ov_raw_seek(&oggStream,0); // \todo defective?
			int res=ov_time_seek(&oggStream,0);

			if(res!=0)
			{
				debug::error(dbgAudio,"Could not rewind ogg stream: %s",getOggErrorString(res).c_str());
			}
/*			else
			{
				debug::info(dbgAudio,"Ogg stream rewinded.");
		}*/
/*
			ov_clear(&oggStream);
			file=0;

			vorbisInfo=0;

			if(!(this->file=fopen(filename.c_str(),"rb")))
			{
				throw std::string("Could not reopen ogg file.");
			}

			int res=ov_open(this->file,&oggStream,0,0);

			if(res<0)
			{
				throw std::string("Could not read Ogg stream:")+getOggErrorString(res);
			}

			vorbisInfo=ov_info(&oggStream,-1);

//			format=(vorbisInfo->channels==1 ? AL_FORMAT_MONO16 : AL_FORMAT_STEREO16);
//			frequency=vorbisInfo->rate;
//			size=(vorbisInfo->channels*2)*ov_pcm_total(&oggStream,-1);
*/
//			//debug::removeMessage();
		}

		char *OggFileLoader::read(ALsizei *bufferSize,bool loop)
		{
			//debug::addMessage("OggFileLoader::read()");

			ALsizei size;

//			debug::debug(dbgAudio,"b01");
			if(bufferSize)
			{
				size=*bufferSize;
			}
			else
			{
				size=this->size;
			}

//			debug::debug(dbgAudio,"b02");
			if(size!=this->bufferSize)
			{
				freeBuffer();
			}
//			debug::debug(dbgAudio,"b03");

			if(!this->buffer)
			{
				this->buffer=new char[size/sizeof(char)];
				this->bufferSize=size;
			}

//			debug::debug(dbgAudio,"b04");
			ALsizei read=0;
			int bitstream=0;

			while(read<size)
			{
//				debug::debug(dbgAudio,"b05: %d,%d,%d",read,size-read,bitstream);
				int res=ov_read(&oggStream,buffer+read,size-read,0,2,1,&bitstream);
//				debug::debug(dbgAudio,"b06");

				if(res>0)
				{
//					debug::debug(dbgAudio,"b07");
					read+=res;
				}
				else if(res<0)
				{
//					debug::debug(dbgAudio,"b08");
					ov_clear(&oggStream);
					throw std::string("Could not decode Ogg stream: ")+getOggErrorString(res);
				}
				else
				{
//					debug::debug(dbgAudio,"b09");
					if(loop)
					{
//						debug::debug(dbgAudio,"b10");
						rewind();
					}
					else
					{
						break;
					}
				}
			}

//			debug::debug(dbgAudio,"b11");
			if(bufferSize)
			{
				*bufferSize=read;
			}

//			debug::debug(dbgAudio,"b12");
			//debug::removeMessage();

			return buffer;
		}

		FileLoader *getFileLoader(std::string filename)
		{
//			//debug::addMessage("getFileLoader()");

			FileLoader *loader=0;
			std::string s=filename;

			if(s.compare(s.size()-4,4,".ogg")==0)
			{
				loader=new OggFileLoader(0,filename);
			}
			else
			{
				throw std::string("Unknown file type - \""+filename+"\"");
			}

//			//debug::removeMessage();

			return(loader);
		}
	};
};
