/**
 * MySDL - A multimedia framework based on SDL and OpenGL
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://www.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "sources.h"

namespace mysdl
{
	namespace audio
	{
		void SoundHandler::soundInit(Sound *sound)
		{
		}

		void SoundHandler::soundUpdate(Sound *sound,GLfloat ticks)
		{
		}

		void SoundHandler::soundStreamed(Sound *sound)
		{
		}

		void SoundHandler::soundDone(Sound *sound)
		{
		}

		Source::Source(SourcePool *owner,ALuint sourceName) :
			owner(owner),state(ssInitialized),priority(0),sound(0),stamp(0),sourceName(sourceName)
		{
		}

		Source::~Source()
		{
			if((state==ssPlaying)||(state==ssPaused))
			{
				//...do something?
			}

			if(sound)
			{
				delete sound;
			}

			if(sourceName)
			{
				alDeleteSources(1,&sourceName);
			}

			/** Does NOT remove itself from its parent's source queue (as it doesn't have direct accesss).
			 *  Therefore, sources must ONLY be deleted by their owning SourcePools, which can take care
			 *  of keeping their queues up to date. */
		}

		/**
		 * Deletes the sound without stopping the source. Use this for manually queuing
		 * multiple sounds in one source.
		 */
// 		void Source::dropSound()
// 		{
// 			if(sound)
// 			{
// 				/*Sound *oldSound=this->sound;
// 				sound=0;
// 				delete oldSound;*/
// 				sound->ignoreStops=true;
// 				sound=0;
// 			}
// 		}

		/**
		 * Compares two sources, sorting them by priority. If one of the sources
		 * has no sample attached, it is considered to be greater. Otherwise,
		 * the source with the lower priority is considered to be greater. This
		 * results in an ordering where the top source is the first to be used
		 * for a new sound.
		 */
		bool Source::operator<(const Source &source) const
		{
			if((stamp==0)&&(source.stamp>0))
			{
				return false;
			}
			else if(source.stamp==0)
			{
				return true;
			}
			else
			{
				if(priority==source.priority)
				{
					return stamp>source.stamp;
				}
				else
				{
					return priority>source.priority;
				}
			}
		}

		SourcePool::SourcePool(int size) :
			sources(),sounds(),deleteSounds(),stampCount(0),volume(1.0)
		{
			if(size)
			{
				resize(size);
			}
		}

		SourcePool::~SourcePool()
		{
			// \todo stop all playing sounds
			deleteExpiredSounds();

			while(!sources.empty())
			{
				delete sources.top();
				sources.pop();
			}
		}

		void SourcePool::deleteExpiredSounds()
		{
			for(std::list<Sound *>::iterator itr=deleteSounds.begin(); itr!=deleteSounds.end(); itr++)
			{
				delete *itr;
			}

			deleteSounds.clear();
		}

		int SourcePool::getSize()
		{
			return sources.size();
		}

		void SourcePool::resize(unsigned int size)
		{
			while(sources.size()>0)
			{
				delete sources.top();
				sources.pop();
			}

			ALuint *sourceNames=new ALuint[size];

			if(!sourceNames)
			{
				throw std::string("Could not create source name array.");
			}
			else
			{
				alGenSources(size,sourceNames);

				// check error

				for(ALuint i=0; i<size; i++)
				{
					sources.push(new Source(this,sourceNames[i]));
				}
			}

			delete sourceNames;
		}

		/**
		 * Warning: This removes the source from the owning SourcePool, since it
		 * must be reinserted at a new position in the priority queue. The caller
		 * MUST therefore call source->owner->push_back(source), otherwise the
		 * requested source will get lost!
		 */
		Source *SourcePool::requestSource(int priority)
		{
			Source *source=sources.top();

			if(source->sound)
			{
				if(source->priority>priority)
				{
					source=0;
				}
				else
				{
					alSourceStop(source->sourceName);
					alSourcei(source->sourceName,AL_BUFFER,AL_NONE);
					delete source->sound;
					source->sound=0;
					source->stamp=0;
				}
			}

			if(source)
			{
				sources.pop();
			}

			return(source);
		}
		
		Sound *SourcePool::play(Sample *sample,const SoundProperties *properties,SoundHandler *handler,int priority)
		{
			if(priority<0)
			{
				priority=sample->priority;
			}

			Source *source=requestSource(priority);

			if(!source)
			{
				debug::info(dbgAudio,"No source availabie, what to do??");
				return 0;
			}
		
			Sound *sound=0;

			if(CachedSample *cachedSample=dynamic_cast<CachedSample *>(sample))
			{
				sound=new CachedSound(this,cachedSample,handler);
			}
			else if(StreamedSample *streamedSample=dynamic_cast<StreamedSample *>(sample))
			{
				sound=new StreamedSound(this,streamedSample,handler);
			}
			else
			{
				throw std::string("Unknown sample class.");
			}

			stampCount++;

			source->sound=sound;
			source->priority=priority;
			source->stamp=stampCount;
			sound->source=source;

			if(properties)
			{
				sound->properties=*properties;
			}
			else
			{
				sound->properties=sample->properties;
			}

			sources.push(source);
			sound->play();
			
			return sound;
		}

		void SourcePool::perform(GLfloat ticks)
		{
			for(std::list<Sound *>::iterator itr=sounds.begin(); itr!=sounds.end(); itr++)
			{
				(*itr)->perform(ticks);
			}

			deleteExpiredSounds();
		}

		void SourcePool::stopAll()
		{
			for(std::list<Sound *>::iterator itr=sounds.begin(); itr!=sounds.end(); itr++)
			{
				(*itr)->stop();
			}

			deleteExpiredSounds();
		}

		void SourcePool::setPlayingVolumes(ALfloat volume)
		{
			setVolume(volume);

			for(std::list<Sound *>::iterator itr=sounds.begin(); itr!=sounds.end(); itr++)
			{
				(*itr)->setVolume((*itr)->properties.volume);
			}
		}

		void SourcePool::setVolume(ALfloat volume) // Note: this will NOT affect sounds which are already playing (do we want this?)
		{
			this->volume=volume;
		}

		ALfloat SourcePool::getVolume()
		{
			return(volume);
		}

		StreamedSound::StreamedSound(SourcePool *owner,const StreamedSample *sample,SoundHandler *handler) :
			Sound(owner,sample,handler),
			loader(0),bufferSize(sample->bufferSize),fileComplete(false)
		{
			buffers[0]=buffers[1]=0;

			try
			{
				loader=getFileLoader(sample->filename);
			}
			catch(std::string ex)
			{
				if(loader)
				{
					delete loader;
					loader=0;
				}

				throw std::string("Error loading file \"")+sample->filename+std::string("\": ")+ex;
			}

			alGenBuffers(2,buffers);
			readBuffer(buffers[0]);
			readBuffer(buffers[1]);
		}

		StreamedSound::~StreamedSound()
		{
			if(buffers[0])
			{
				alDeleteBuffers(2,buffers);
				buffers[0]=buffers[1]=0;
			}

			if(loader)
			{
				delete loader;
				loader=0;
			}
		}

		void StreamedSound::readBuffer(ALuint bufferName)
		{
			ALsizei size=bufferSize;
			char *data=loader->read(&size,properties.loop);

			if(size)
			{
				alBufferData(bufferName,loader->getFormat(),data,size,loader->getFrequency());
			}
			else
			{
				fileComplete=true;
				handler->soundStreamed(this);
			}
		}

		void StreamedSound::stream()
		{
			ALint processed;

			for(alGetSourcei(source->sourceName,AL_BUFFERS_PROCESSED,&processed); processed>0; processed--)
			{
/*				if(queueCount>0)
				{
					queueCount--;
					continue;
				}*/

				//printf("streaming...\n"); fflush(stdout);
				ALuint buffer;

				alSourceUnqueueBuffers(source->sourceName,1,&buffer);
				readBuffer(buffer);

				if(!fileComplete)
				{
					alSourceQueueBuffers(source->sourceName,1,&buffer);
				}
			}

			ALint hasStopped;
			alGetSourcei(source->sourceName,AL_SOURCE_STATE,&hasStopped);

			if(hasStopped==AL_STOPPED)
			{
				if(fileComplete)
				{
					source->state=ssStopped;
				}
				else
				{
					debug::warning(dbgAudio,"Stream %p was interrupted.",this);
					alSourcePlay(source->sourceName);
				}
			}
		}

		void StreamedSound::play()
		{
			alSourceQueueBuffers(source->sourceName,1,&buffers[0]); // \todo move to constructor, same goes for cached sounds
			alSourceQueueBuffers(source->sourceName,1,&buffers[1]);
			
			alSourcef(source->sourceName,AL_GAIN,owner->volume*properties.volume);
			alSourcef(source->sourceName,AL_PITCH,properties.pitch);
			alSourcef(source->sourceName,AL_ROLLOFF_FACTOR,1.0);
			alSourcefv(source->sourceName,AL_POSITION,&(properties.position.x)); // \todo move to common location

			alSourcei(source->sourceName,AL_LOOPING,AL_FALSE);
			Sound::play();
		}

		void StreamedSound::perform(GLfloat ticks)
		{
			stream();
			Sound::perform(ticks);
		}

		CachedSound::CachedSound(SourcePool *owner,const CachedSample *sample,SoundHandler *handler) :
			Sound(owner,sample,handler)
		{
		}

		void CachedSound::play()
		{
			const CachedSample *sample=(CachedSample *)(this->sample);
			ALuint bufferName=sample->bufferName;

			ALfloat pos[3]={properties.position.x,properties.position.y,properties.position.z};
			alSourcef(source->sourceName,AL_GAIN,owner->volume*properties.volume);
			alSourcef(source->sourceName,AL_PITCH,properties.pitch);

			alSourcei(source->sourceName,AL_LOOPING,properties.loop);
			alSourceQueueBuffers(source->sourceName,1,&(bufferName));
			alSourcefv(source->sourceName,AL_POSITION,pos); // \todo move to common location

			Sound::play();
		}

		Sound::Sound(SourcePool *owner,const Sample *sample,SoundHandler *handler) :
			owner(owner),sample(sample),handler(handler),source(0),properties()
		{
			owner->sounds.push_back(this);

			if(handler)
			{
				handler->soundInit(this);
			}
		}

		Sound::~Sound()
		{
			if(handler)
			{
				handler->soundDone(this);
			}

			if(source->sound==this)
			{
				alSourceStop(source->sourceName);
				alSourcei(source->sourceName,AL_BUFFER,AL_NONE);
				source->state=ssStopped;
				source->sound=0;
			}

			owner->sounds.remove(this);
		}

		void Sound::setVolume(ALfloat volume)
		{
			properties.volume=volume;
			alSourcef(source->sourceName,AL_GAIN,owner->volume*volume);
		}

		void Sound::perform(GLfloat ticks)
		{
			if(handler)
			{
				handler->soundUpdate(this,ticks);
			}

			ALint hasStopped;
			alGetSourcei(source->sourceName,AL_SOURCE_STATE,&hasStopped);

			if(hasStopped==AL_STOPPED)
			{
				source->state=ssStopped;

				alSourcei(source->sourceName,AL_BUFFER,AL_NONE);
				owner->deleteSounds.push_back(this);
			}
		}

		/**
		 * This method doesn't make sense by itself. Derived classes must override it by
		 * queuing their initial audio buffers before calling this inherited method.
		 */
		void Sound::play()
		{
			alSourcePlay(source->sourceName); // \todo check errors
			source->state=ssPlaying;
		}

		void Sound::stop()
		{
			alSourceStop(source->sourceName);
			alSourcei(source->sourceName,AL_BUFFER,AL_NONE);
			source->state=ssStopped;
			owner->deleteSounds.push_back(this);
		}

		Source *Sound::getSource()
		{
			return source;
		}
	};
};

/** \todo don't check generated source ids for 0!! */
