/**
 * MySDL - A multimedia framework based on SDL and OpenGL
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://www.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <fstream>
#include <iostream>

#include "controls.h"
#include "debug.h"
#include "mysdl.h"
#include "tools.h"

namespace mysdl
{
	static const std::string XML_TRIGGER="trigger";
	static const std::string XML_QUEUE="queue";
	static const std::string XML_DATA="data";
	static const std::string XML_DATASTRING="dataString";
	static const std::string XML_AXIS="axis";
	static const std::string XML_CODE="code";
	static const std::string XML_NAME="name";
	static const std::string XML_DESCRIPTION="description";
	static const std::string XML_MIN="min";
	static const std::string XML_MAX="max";
	static const std::string XML_DEFAULTKEYSTATE="defaultKeyState";
	static const std::string XML_DIRECTION="direction";
	static const std::string XML_KEY="key";
	static const std::string XML_BUTTON="button";
	static const std::string XML_JOY="joy";
	static const std::string XML_VALUE="value";
	static const std::string XML_STATE="state";
	static const std::string XML_KEYTRANSLATOR="keyTranslator";
	static const std::string XML_JOYBUTTONTRANSLATOR="joyButtonTranslator";
	static const std::string XML_JOYAXISTRANSLATOR="joyAxisTranslator";
	static const std::string XML_MOUSEBUTTONTRANSLATOR="mouseButtonTranslator";
	static const std::string XML_MOUSEAXISTRANSLATOR="mouseAxisTranslator";

	template<> std::string toStr<KeyState>(KeyState var)
	{
		switch(var)
		{
			case ctrlDown:		return("down");
			case ctrlUp:			return("up");
			case ctrlPressed:	return("pressed");
			default:					return("default");
		}
 	}

	template<> KeyState strTo<KeyState>(std::string var)
	{
		if(var=="down")
		{
			return(ctrlDown);
		}
		else if(var=="up")
		{
			return(ctrlUp);
		}
		else if(var=="pressed")
		{
			return(ctrlPressed);
		}
		else
		{
			return(ctrlDefault);
		}
	}

	template<> std::string toStr<SDLKey>(SDLKey var)
	{
		return(getKeyNameMap()[var]);
	}

	template<> SDLKey strTo<SDLKey>(std::string var)
	{
		//debug::addMessage("strTo<SDLKey>()");

		static std::map<std::string,SDLKey> map=getNameKeyMap();

		std::map<std::string,SDLKey>::const_iterator itr=map.find(var);

		if(itr==map.end())
		{
			//debug::removeMessage();
			throw std::string(var+" is no valid key identifier.");
		}
		else
		{
			//debug::removeMessage();
			return(itr->second);
		}
	}

	ControlQueue queue;

	Evaluator::Evaluator() :
		enabled(true)
	{
	}

	Evaluator::~Evaluator()
	{
	}

	void Evaluator::perform(GLfloat ticks,ControlQueue *queue)
	{
	}

	void Evaluator::flush()
	{
	}

	ControlQueue::ControlQueue() :
		eventQueue(),evaluators(),activateTrigger(0),deactivateTrigger(0)
	{
	}

	ControlQueue::~ControlQueue()
	{
		//debug::addMessage("ControlQueue::~ControlQueue()");

		eventQueue.clear();
		evaluators.clear();

		//debug::removeMessage();
	}

	void ControlQueue::operator<<(const SDL_Event &event)
	{
		//debug::addMessage("ControlQueue::operator<<(SDL_Event&)");

		for(std::list<Evaluator *>::iterator itr=evaluators.begin(); itr!=evaluators.end(); itr++)
		{
			if(((*itr)->enabled)&&((*itr)->evaluate(event,this)))
			{
				break;
			}
		}

		//debug::removeMessage();
	}

	void ControlQueue::operator<<(Trigger *trigger)
	{
		//debug::addMessage("ControlQueue::operator<<(Trigger *)");
		eventQueue.push_back(trigger);
		//debug::removeMessage();
	}

	void ControlQueue::operator>>(Trigger *&result)
	{
		//debug::addMessage("ControlQueue::operator>>()");

		if(isEmpty())
		{
			result=0;
		}
		else
		{
			result=eventQueue.front();
			eventQueue.pop_front();
		}

		//debug::removeMessage();
	}

	void ControlQueue::perform(GLfloat ticks)
	{
		//debug::addMessage("ControlQueue::perform()");

		for(std::list<Evaluator *>::iterator itr=evaluators.begin(); itr!=evaluators.end(); itr++)
		{
			if((*itr)->enabled)
			{
				(*itr)->perform(ticks,this);
			}
		}

		//debug::removeMessage();
	}

	void ControlQueue::flush()
	{
		//debug::addMessage("ControlQueue::flush()");

		for(std::list<Evaluator *>::iterator itr=evaluators.begin(); itr!=evaluators.end(); itr++)
		{
			(*itr)->flush();
		}

		//debug::removeMessage();
	}

	bool ControlQueue::isEmpty()
	{
		return(eventQueue.empty());
	}

	bool ControlQueue::readEvents()
	{
		//debug::addMessage("ControlQueue::readEvents()");

		bool noQuit=true;
		SDL_Event event;

		while(SDL_PollEvent(&event))
		{
			switch(event.type)
			{
				case SDL_QUIT:
				{
					noQuit=false;
					break;
				}
				
				case SDL_ACTIVEEVENT:
				{
					if((event.active.state & SDL_APPINPUTFOCUS)||
						 (event.active.state & SDL_APPACTIVE))
					{
						if(event.active.gain)
						{
							if(activateTrigger)
							{
								(*this) << activateTrigger;
							}
						}
						else
						{
							if(deactivateTrigger)
							{
								(*this) << deactivateTrigger;
							}
						}
					}
				}

				default:
				{
					(*this) << event;
					break;
				}
			}
		}

		//debug::removeMessage();
		return(noQuit);
	}

	Translator::Translator(Trigger *trigger) :
		trigger(trigger)
	{
		//debug::addMessage("Translator::Translator()");
		trigger->translators.push_back(this);
		//debug::removeMessage();
	}

	Translator::~Translator()
	{
		//debug::addMessage("Translator::~Translator()");
		trigger->translators.remove(this);
		//debug::removeMessage();
	}

	Trigger::Trigger(InputEvaluator *evaluator,ControlId id,const std::string &name,const std::string &humanName,KeyState defaultKeyState) :
		NamedResource<InputEvaluator>(name),
		id(id),humanName(humanName),data(0),dataString(),translators(),digital(digital),max(1.0),defaultKeyState(defaultKeyState),queue(true),
		enabled(true),value(0)
	{
		//debug::addMessage("Trigger::Trigger()");
		(*evaluator)+=this;
		//debug::removeMessage();
	}

	Trigger::Trigger(InputEvaluator *evaluator,ControlId id,const std::string &name,const std::string &humanName,GLfloat max) :
		NamedResource<InputEvaluator>(name),
		id(id),humanName(humanName),data(0),dataString(),translators(),digital(false),max(max),defaultKeyState(ctrlPressed),queue(false),
		enabled(true),value(0)
	{
		//debug::addMessage("Trigger::Trigger()");
		(*evaluator)+=this;
		//debug::removeMessage();
	}

	Trigger::~Trigger()
	{
		//debug::addMessage("Trigger::~Trigger()");
		clearTranslators();
		//debug::removeMessage();
	}

	void Trigger::clearTranslators()
	{
		//debug::addMessage("Trigger::clearTranslators()");

		while(!(translators.empty()))
		{
			delete(translators.front());
		}

		//debug::removeMessage();
	}

	void Trigger::clearKeyTranslators(GLfloat value)
	{
		//debug::addMessage("Trigger::clearKeyTranslators()");

		std::list<KeyTranslator *> deleteList;

		for(std::list<Translator *>::iterator itr=translators.begin(); itr!=translators.end(); itr++)
		{
			if(KeyTranslator *kt=dynamic_cast<KeyTranslator *>(*itr))
			{
				if((value==0.0)||(value==kt->value))
				{
					deleteList.push_back(kt);
				}
			}
		}

		for(std::list<KeyTranslator *>::iterator itr=deleteList.begin(); itr!=deleteList.end(); itr++)
		{
			delete *itr;
		}

		//debug::removeMessage();
	}

	void Trigger::clearJoyButtonTranslators(GLfloat value)
	{
		//debug::addMessage("Trigger::clearJoyButtonTranslators()");

		std::list<JoyButtonTranslator *> deleteList;

		for(std::list<Translator *>::iterator itr=translators.begin(); itr!=translators.end(); itr++)
		{
			if(JoyButtonTranslator *jbt=dynamic_cast<JoyButtonTranslator *>(*itr))
			{
				if((value==0.0)||(value==jbt->value))
				{
					deleteList.push_back(jbt);
				}
			}
		}

		for(std::list<JoyButtonTranslator *>::iterator itr=deleteList.begin(); itr!=deleteList.end(); itr++)
		{
			delete *itr;
		}

		//debug::removeMessage();
	}

	void Trigger::clearJoyAxisTranslators(short int direction)
	{
		//debug::addMessage("Trigger::clearJoyAxisTranslators()");

		std::list<JoyAxisTranslator *> deleteList;

		for(std::list<Translator *>::iterator itr=translators.begin(); itr!=translators.end(); itr++)
		{
			if(JoyAxisTranslator *jat=dynamic_cast<JoyAxisTranslator *>(*itr))
			{
				if((direction==0.0)||(direction==jat->direction))
				{
					deleteList.push_back(jat);
				}
			}
		}

		for(std::list<JoyAxisTranslator *>::iterator itr=deleteList.begin(); itr!=deleteList.end(); itr++)
		{
			delete *itr;
		}

		//debug::removeMessage();
	}

	void Trigger::clearMouseButtonTranslators(GLfloat value)
	{
		//debug::addMessage("Trigger::clearMouseButtonTranslators()");

		std::list<MouseButtonTranslator *> deleteList;

		for(std::list<Translator *>::iterator itr=translators.begin(); itr!=translators.end(); itr++)
		{
			if(MouseButtonTranslator *mbt=dynamic_cast<MouseButtonTranslator *>(*itr))
			{
				if((value==0.0)||(value==mbt->value))
				{
					deleteList.push_back(mbt);
				}
			}
		}

		for(std::list<MouseButtonTranslator *>::iterator itr=deleteList.begin(); itr!=deleteList.end(); itr++)
		{
			delete *itr;
		}

		//debug::removeMessage();
	}

	void Trigger::clearMouseAxisTranslators(short int direction)
	{
		//debug::addMessage("Trigger::clearMouseAxisTranslators()");

		std::list<MouseAxisTranslator *> deleteList;

		for(std::list<Translator *>::iterator itr=translators.begin(); itr!=translators.end(); itr++)
		{
			if(MouseAxisTranslator *mat=dynamic_cast<MouseAxisTranslator *>(*itr))
			{
				if((direction==0.0)||(direction==mat->direction))
				{
					deleteList.push_back(mat);
				}
			}
		}

		for(std::list<MouseAxisTranslator *>::iterator itr=deleteList.begin(); itr!=deleteList.end(); itr++)
		{
			delete *itr;
		}

		//debug::removeMessage();
	}

	std::string Trigger::keysToString(GLfloat value)
	{
		//debug::addMessage("Trigger::keysToString()");

		std::string s;

		for(std::list<Translator *>::iterator itr=translators.begin(); itr!=translators.end(); itr++)
		{
			if(KeyTranslator *kt=dynamic_cast<KeyTranslator *>(*itr))
			{
				if((value==0.0)||(kt->value==value))
				{
					if(!(s.empty()))
					{
						s+=";";
					}

					s+=toStr(kt->key);
				}
			}
		}

		//debug::removeMessage();
		return(s);
	}

	std::string Trigger::joyToString(short int direction,GLfloat value)
	{
		//debug::addMessage("Trigger::joyToString()");

		std::string s1=joyAxisToString(direction);
		std::string s2=joyButtonsToString(value);

		if(s1.size()==0)
		{
			return(s2);
		}
		else
		{
			if(s2.size()==0)
			{
				return(s1);
			}
			else
			{
				return(s1+";"+s2);
			}
		}

		//debug::removeMessage();
	}

	std::string Trigger::joyButtonsToString(GLfloat value)
	{
		//debug::addMessage("Trigger::joyButtonsToString()");

		std::string s;

		for(std::list<Translator *>::iterator itr=translators.begin(); itr!=translators.end(); itr++)
		{
			if(JoyButtonTranslator *jbt=dynamic_cast<JoyButtonTranslator *>(*itr))
			{
				if((value==0.0)||(jbt->value==value))
				{
					if(!(s.empty()))
					{
						s+=";";
					}

					s+="Joy "+toStr<int>(jbt->joy)+"; Btn "+toStr<int>(jbt->button);
				}
			}
		}

		//debug::removeMessage();
		return(s);
	}

	std::string Trigger::joyAxisToString(short int direction)
	{
		//debug::addMessage("Trigger::joyAxisToString()");

		std::string s;

		for(std::list<Translator *>::iterator itr=translators.begin(); itr!=translators.end(); itr++)
		{
			if(JoyAxisTranslator *jat=dynamic_cast<JoyAxisTranslator *>(*itr))
			{
				if((direction==0)||(jat->direction==direction))
				{
					if(!(s.empty()))
					{
						s+=";";
					}

					s+="Joy "+toStr<int>(jat->joy)+"; Axis "+toStr<int>(jat->axis)+(jat->direction ? (jat->direction>0 ? " +" : " -") : "");
				}
			}
		}

		//debug::removeMessage();
		return(s);
	}

	Trigger *Trigger::clone(InputEvaluator *evaluator)
	{
		//debug::addMessage("Trigger::clone()");

		Trigger *trigger=new Trigger(evaluator,id,name,humanName);
		trigger->data=data;
		trigger->dataString=dataString;
		trigger->digital=digital;
		trigger->max=max;
		trigger->enabled=enabled;
		trigger->defaultKeyState=defaultKeyState;
		trigger->queue=queue;

		for(std::list<Translator *>::iterator itr=translators.begin(); itr!=translators.end(); itr++)
		{
			(*itr)->clone(trigger);
		}

		//debug::removeMessage();
		return(trigger);
	}

	bool Trigger::isActive()
	{
		return(value!=0);
	}

	void Trigger::trigger(ControlQueue *queue)
	{
		//debug::addMessage("Trigger::trigger(ControlQueue*)");

		if(this->queue)
		{
			(*queue) << this;
		}

		//debug::removeMessage();
	}

	void Trigger::trigger(ControlQueue *queue,Translator *translator,GLfloat value)
	{
		//debug::addMessage("Trigger::trigger(ControlQueue*,Translator*,GLfloat)");

		this->value=value;

		if(this->queue)
		{
			(*queue) << this;
		}

		//debug::removeMessage();
	}

	void Trigger::setValue(ControlQueue *queue,Translator *translator,GLfloat value)
	{
		//debug::addMessage("Trigger::setValue(ControlQueue*,Translator*,GLfloat)");

		if(value!=this->value)
		{
			this->value=value;

			if(this->queue)
			{
				(*queue) << this;
			}
		}

		//debug::removeMessage();
	}

	void Trigger::reset()
	{
		value=0;
	}

	KeyTranslator::KeyTranslator(Trigger *trigger,SDLKey key,KeyState state,GLfloat value) :
		Translator(trigger),key(key),state(state),value(value)
	{
		//debug::addMessage("KeyTranslator::KeyTranslator()");

		if(state==ctrlDefault)
		{
			this->state=trigger->defaultKeyState;
		}

		trigger->owner->translatorsKey.push_back(this);

		//debug::removeMessage();
	}

	KeyTranslator::~KeyTranslator()
	{
		//debug::addMessage("KeyTranslator::~KeyTranslator()");
		trigger->owner->translatorsKey.remove(this);
		//debug::removeMessage();
	}

	xml::Node *KeyTranslator::buildTranslatorTree()
	{
		//debug::addMessage("KeyTranslator::buildTranslatorTree()");

		xml::Node *root=new xml::Node();

		root->name=XML_KEYTRANSLATOR;
		(*root)[XML_VALUE]=toStr<GLfloat>(value);
		(*root)[XML_KEY]=toStr(key);

		if(state!=trigger->defaultKeyState)
		{
			(*root)[XML_STATE]=toStr(state);
		}

		//debug::removeMessage();
		return(root);
	}

	Translator *KeyTranslator::clone(Trigger *trigger)
	{
		return(new KeyTranslator(trigger,key,state,value));
	}

	bool KeyTranslator::translate(const SDL_Event &event,ControlQueue *queue)
	{
		//debug::addMessage("KeyTranslator::translate()");

		if(state==ctrlPressed)
		{
			if(event.type==SDL_KEYDOWN)
			{
				trigger->owner->activeTriggers.insert(trigger);
				trigger->value=value;
			}
			else
			{
				if(trigger->value==value)
				{
					trigger->owner->activeTriggers.erase(trigger);
					trigger->trigger(queue,this,0);
				}
			}
		}
		else
		{
			trigger->trigger(queue,this,value);
		}

		//debug::removeMessage();
		return(true);
	}

	JoyTranslator::JoyTranslator(Trigger *trigger,Sint16 joy) :
		Translator(trigger),joy(joy)
	{
	}

	JoyButtonTranslator::JoyButtonTranslator(Trigger *trigger,Sint16 joy,Sint16 button,KeyState state,GLfloat value) :
		JoyTranslator(trigger,joy),
		button(button),state(state),value(value)
	{
		//debug::addMessage("JoyButtonTranslator::JoyButtonTranslator()");

		if(state==ctrlDefault)
		{
			this->state=trigger->defaultKeyState;
		}

		trigger->owner->translatorsJoyButton.push_back(this);

		//debug::removeMessage();
	}

	JoyButtonTranslator::~JoyButtonTranslator()
	{
		//debug::addMessage("JoyButtonTranslator::~JoyButtonTranslator()");
		trigger->owner->translatorsJoyButton.remove(this);
		//debug::removeMessage();
	}

	xml::Node *JoyButtonTranslator::buildTranslatorTree()
	{
		//debug::addMessage("JoyButtonTranslator::buildTranslatorTree()");

		xml::Node *root=new xml::Node();

		root->name=XML_JOYBUTTONTRANSLATOR;
		(*root)[XML_VALUE]=toStr<GLfloat>(value);
		(*root)[XML_JOY]=toStr<int>(joy);
		(*root)[XML_BUTTON]=toStr<int>(button);

		if(state!=trigger->defaultKeyState)
		{
			(*root)[XML_STATE]=toStr(state);
		}

		//debug::removeMessage();
		return(root);
	}

	Translator *JoyButtonTranslator::clone(Trigger *trigger)
	{
		return(new JoyButtonTranslator(trigger,joy,button,state,value));
	}

	bool JoyButtonTranslator::translate(const SDL_Event &event,ControlQueue *queue)
	{
		//debug::addMessage("JoyButtonTranslator::translate()");

		if(state==ctrlPressed)
		{
			if(event.type==SDL_JOYBUTTONDOWN)
			{
				trigger->owner->activeTriggers.insert(trigger);
				trigger->value=value;
			}
			else
			{
				if(trigger->value==value) {
				trigger->owner->activeTriggers.erase(trigger);
				trigger->trigger(queue,this,0);
				}
			}
		}
		else
		{
			trigger->trigger(queue,this,value);
		}

		//debug::removeMessage();
		return(true);
	}

	JoyAxisTranslator::JoyAxisTranslator(Trigger *trigger,Sint16 joy,Uint8 axis,Sint8 direction,Sint16 minPhysical,Sint16 maxPhysical) :
		JoyTranslator(trigger,joy),
		axis(axis),direction(direction),minPhysical(minPhysical),maxPhysical(maxPhysical)
	{
		//debug::addMessage("JoyAxisTranslator::JoyAxisTranslator()");
		trigger->owner->translatorsJoyAxis.push_back(this);
		//debug::removeMessage();
	}

	JoyAxisTranslator::~JoyAxisTranslator()
	{
		//debug::addMessage("JoyAxisTranslator::~JoyAxisTranslator()");
		trigger->owner->translatorsJoyAxis.remove(this);
		//debug::removeMessage();
	}

	xml::Node *JoyAxisTranslator::buildTranslatorTree()
	{
		//debug::addMessage("JoyAxisTranslator::buildTranslatorTree()");

		xml::Node *root=new xml::Node();

		root->name=XML_JOYAXISTRANSLATOR;
		(*root)[XML_JOY]=toStr<int>(joy);
		(*root)[XML_AXIS]=toStr<int>(axis);

		if(direction!=0)
		{
			(*root)[XML_DIRECTION]=toStr<int>(direction);
		}

		(*root)[XML_MIN]=toStr(minPhysical);
		(*root)[XML_MAX]=toStr(maxPhysical);

		//debug::removeMessage();
		return(root);
	}

	Translator *JoyAxisTranslator::clone(Trigger *trigger)
	{
		return(new JoyAxisTranslator(trigger,joy,axis,direction,minPhysical,maxPhysical));
	}

	bool JoyAxisTranslator::translate(const SDL_Event &event,ControlQueue *queue)
	{
		//debug::addMessage("JoyAxisTranslator::translate()");

		if(trigger->digital)
		{
			trigger->setValue(queue,this,abs(event.jaxis.value)>minPhysical ? trigger->max : 0.0);
			//debug::removeMessage();
			//mysdl::debug::debug(dbgControls,"Digital; value for axis %d is now %.2f.",trigger->id,trigger->value);
			return true;
		}
		else
		{
			short int direction=(event.jaxis.value==0 ? 0 : event.jaxis.value>0 ? +1 : -1);

			if((this->direction==0)||(direction==0)||(this->direction==direction))
			{
				//Sint16 value=abs(event.jaxis.value);
				Sint32 minPhysical=this->minPhysical;
				Sint32 maxPhysical=this->maxPhysical;
				Sint32 value=abs(event.jaxis.value);

				if(value<minPhysical)
				{
					trigger->setValue(queue,this,0.0);
				}
				else if(value>=maxPhysical)
				{
					trigger->setValue(queue,this,((GLfloat)direction)*(GLfloat)trigger->max);
				}
				else
				{
					trigger->setValue(queue,this,((GLfloat)direction)*(GLfloat)trigger->max*(((GLfloat)value)-minPhysical)/((GLfloat)maxPhysical-(GLfloat)minPhysical));
				}

				//mysdl::debug::debug(dbgControls,"Analog; value for axis %d is now %.2f.",trigger->id,trigger->value);
				//debug::removeMessage();
				return true;
			}
		}

		//debug::removeMessage();
		return false;
	}

	MouseButtonTranslator::MouseButtonTranslator(Trigger *trigger,Uint8 button,KeyState state,GLfloat value) :
		Translator(trigger),
		button(button),state(state),value(value)
	{
		//debug::addMessage("MouseButtonTranslator::MouseButtonTranslator()");

		if(state==ctrlDefault)
		{
			this->state=trigger->defaultKeyState;
		}

		trigger->owner->translatorsMouseButton.push_back(this);

		//debug::removeMessage();
	}

	MouseButtonTranslator::~MouseButtonTranslator()
	{
		//debug::addMessage("MouseButtonTranslator::~MouseButtonTranslator()");
		trigger->owner->translatorsMouseButton.remove(this);
		//debug::removeMessage();
	}

	xml::Node *MouseButtonTranslator::buildTranslatorTree()
	{
		//debug::addMessage("MouseButtonTranslator::buildTranslatorTree()");

		xml::Node *root=new xml::Node();

		root->name=XML_MOUSEBUTTONTRANSLATOR;
		(*root)[XML_VALUE]=toStr<GLfloat>(value);
		(*root)[XML_BUTTON]=toStr<int>(button);

		if(state!=trigger->defaultKeyState)
		{
			(*root)[XML_STATE]=toStr(state);
		}

		//debug::removeMessage();
		return(root);
	}

	Translator *MouseButtonTranslator::clone(Trigger *trigger)
	{
		return(new MouseButtonTranslator(trigger,button,state,value));
	}

	bool MouseButtonTranslator::translate(const SDL_Event &event,ControlQueue *queue)
	{
		//debug::addMessage("MouseButtonTranslator::translate()");

		if(state==ctrlPressed)
		{
			if(event.type==SDL_MOUSEBUTTONDOWN)
			{
				trigger->owner->activeTriggers.insert(trigger);
				trigger->value=value;
			}
			else
			{
				if(trigger->value==value)
				{
					trigger->owner->activeTriggers.erase(trigger);
					trigger->trigger(queue,this,0);
				}
			}
		}
		else
		{
			trigger->trigger(queue,this,value);
		}

		//debug::removeMessage();
		return(true);
	}

	MouseAxisTranslator::MouseAxisTranslator(Trigger *trigger,Uint8 axis,Sint8 direction,Sint16 minPhysical,Sint16 maxPhysical) :
		Translator(trigger),
		axis(axis),direction(direction),minPhysical(minPhysical),maxPhysical(maxPhysical)
	{
		//debug::addMessage("MouseAxisTranslator::MouseAxisTranslator()");
		trigger->owner->translatorsMouseAxis.push_back(this);
		//debug::removeMessage();
	}

	MouseAxisTranslator::~MouseAxisTranslator()
	{
		//debug::addMessage("MouseAxisTranslator::~MouseAxisTranslator()");
		trigger->owner->translatorsMouseAxis.remove(this);
		//debug::removeMessage();
	}

	xml::Node *MouseAxisTranslator::buildTranslatorTree()
	{
		//debug::addMessage("MouseAxisTranslator::buildTranslatorTree()");

		xml::Node *root=new xml::Node();

		root->name=XML_MOUSEAXISTRANSLATOR;
		(*root)[XML_AXIS]=toStr<int>(axis);

		if(direction!=0)
		{
			(*root)[XML_DIRECTION]=toStr<int>(direction);
		}

		(*root)[XML_MIN]=toStr(minPhysical);
		(*root)[XML_MAX]=toStr(maxPhysical);

		//debug::removeMessage();
		return(root);
	}

	Translator *MouseAxisTranslator::clone(Trigger *trigger)
	{
		return(new MouseAxisTranslator(trigger,axis,direction,minPhysical,maxPhysical));
	}

	bool MouseAxisTranslator::translate(const SDL_Event &event,ControlQueue *queue)
	{
		//debug::addMessage("MouseAxisTranslator::translate()");

		Sint16 v=(axis==0 ? event.motion.xrel : event.motion.yrel);
		
		if(trigger->digital)
		{
			trigger->setValue(queue,this,abs(v)>minPhysical ? trigger->max : 0.0);
			//debug::removeMessage();
			return(true);
		}
		else
		{
			short int direction=(v==0 ? 0 : v>0 ? +1 : -1);

			if((this->direction==0)||(direction==0)||(this->direction==direction))
			{
				Sint16 value=abs(v);

				if(value<minPhysical)
				{
					trigger->setValue(queue,this,0.0);
				}
				else if(value>=maxPhysical)
				{
					trigger->setValue(queue,this,((GLfloat)direction)*trigger->max);
				}
				else
				{
					trigger->setValue(queue,this,((GLfloat)direction)*trigger->max*(((GLfloat)value)-minPhysical)/(maxPhysical-minPhysical));
				}

				//debug::removeMessage();
				return(true);
			}
		}
		
		//debug::removeMessage();
		return(false);
	}
























InputEvaluator::InputEvaluator() :
		Evaluator(),NamedResourceBank<Trigger>(),
		/*triggers(),*/translatorsKey(),translatorsJoyButton(),translatorsJoyAxis(),translatorsMouseButton(),translatorsMouseAxis(),activeTriggers()
	{
	}

	InputEvaluator::~InputEvaluator()
	{
		//debug::addMessage("InputEvaluator::~InputEvaluator()");
		clear();
		//debug::removeMessage();
	}

	void InputEvaluator::setDataOnAllTriggers(int data)
	{
		//debug::addMessage("InputEvaluator::setDataOnAllTriggers()");

		for(InputEvaluator::iterator itr=begin(); itr!=end(); itr++)
		{
			(*itr)->data=data;
		}

		//debug::removeMessage();
	}

	void InputEvaluator::clear()
	{
		//debug::addMessage("InputEvaluator::clear()");

		NamedResourceBank<Trigger>::clear();

		translatorsKey.clear();
		translatorsJoyButton.clear();
		translatorsJoyAxis.clear();
		translatorsMouseButton.clear();
		translatorsMouseAxis.clear();
		activeTriggers.clear();

		//debug::removeMessage();
	}

	void InputEvaluator::copyFrom(InputEvaluator *source)
	{
		//debug::addMessage("InputEvaluator::copyFrom()");

		clear();

		for(InputEvaluator::iterator itr=source->begin(); itr!=source->end(); itr++)
		{
			(*itr)->clone(this);
		}

		//debug::removeMessage();
	}

	InputEvaluator *InputEvaluator::clone()
	{
		//debug::addMessage("InputEvaluator::clone()");

		InputEvaluator *eval=new InputEvaluator();
		eval->copyFrom(this);

		//debug::removeMessage();
		return(eval);
	}

	bool InputEvaluator::evaluate(const SDL_Event &event,ControlQueue *queue)
	{
		if(size()==0)
		{
			return(false);
		}

		switch(event.type)
		{
			case SDL_KEYDOWN:
			case SDL_KEYUP:
				return(handleKeyEvent(event,queue));

			case SDL_JOYBUTTONDOWN:
			case SDL_JOYBUTTONUP:
				return(handleJoyButtonEvent(event,queue));

			case SDL_JOYAXISMOTION:
				return(handleJoyAxisEvent(event,queue));

			case SDL_MOUSEBUTTONDOWN:
			case SDL_MOUSEBUTTONUP:
				return(handleMouseButtonEvent(event,queue));

			case SDL_MOUSEMOTION:
				return(handleMouseAxisEvent(event,queue));

			default:
				return(false);
		}
	}

	void InputEvaluator::perform(GLfloat ticks,ControlQueue *queue)
	{
		//debug::addMessage("InputEvaluator::perform()");

		for(std::set<Trigger *>::iterator itr=activeTriggers.begin(); itr!=activeTriggers.end(); itr++)
		{
			(*queue) << *itr;
		}

		//debug::removeMessage();
	}

	void InputEvaluator::flush()
	{
		//debug::addMessage("InputEvaluator::flush()");

		for(std::set<Trigger *>::iterator itr=activeTriggers.begin(); itr!=activeTriggers.end(); itr++)
		{
			(*itr)->reset();
		}

		activeTriggers.clear();

		//debug::removeMessage();
	}

	bool InputEvaluator::handleKeyEvent(const SDL_Event &e,ControlQueue *queue)
	{
		//debug::addMessage("InputEvaluator::handleKeyEvent()");

		bool result=false;

		for(std::list<KeyTranslator *>::iterator itr=translatorsKey.begin(); itr!=translatorsKey.end(); itr++)
		{
			if((*itr)->key==e.key.keysym.sym)
			{
				if((((*itr)->state==ctrlUp)&&(e.type==SDL_KEYUP))||
					 (((*itr)->state==ctrlDown)&&(e.type==SDL_KEYDOWN))||
					 (((*itr)->state==ctrlPressed)))
				{
					result|=(*itr)->translate(e,queue);
				}
			}
		}

		//debug::removeMessage();
		return(result);
	}

	bool InputEvaluator::handleJoyButtonEvent(const SDL_Event &e,ControlQueue *queue)
	{
		//debug::addMessage("InputEvaluator::handleJoyButtonEvent()");

		bool result=false;
		bool translated=false;
		JoyButtonTranslator *fallback=0;
		
		for(std::list<JoyButtonTranslator *>::iterator itr=translatorsJoyButton.begin(); itr!=translatorsJoyButton.end(); itr++)
		{
			if(((*itr)->joy==e.jbutton.which)&&((*itr)->button==e.jbutton.button))
			{
				if((((*itr)->state==ctrlUp)&&(e.type==SDL_JOYBUTTONUP))||
					 (((*itr)->state==ctrlDown)&&(e.type==SDL_JOYBUTTONDOWN))||
					 (((*itr)->state==ctrlPressed)))
				{
					result|=(*itr)->translate(e,queue);
					translated=true;
				}
			}
			else if(!translated)
			{
				if(((*itr)->joy<0)||((*itr)->joy==e.jbutton.which))
				{
					if(((*itr)->button<0)||((*itr)->button==e.jbutton.button))
					{
						if((((*itr)->state==ctrlUp)&&(e.type==SDL_JOYBUTTONUP))||
												 (((*itr)->state==ctrlDown)&&(e.type==SDL_JOYBUTTONDOWN))||
												 (((*itr)->state==ctrlPressed)))
						{
							fallback=*itr;
						}	
					}
				}
			}
		}

		if((!translated)&&(fallback))
		{
			result|=fallback->translate(e,queue);
		}
		
		//debug::removeMessage();
		return(result);
	}

	bool InputEvaluator::handleJoyAxisEvent(const SDL_Event &e,ControlQueue *queue)
	{
		//debug::addMessage("InputEvaluator::handleJoyAxisEvent()");

		bool result=false;

		for(std::list<JoyAxisTranslator *>::iterator itr=translatorsJoyAxis.begin(); itr!=translatorsJoyAxis.end(); itr++)
		{
			if(((*itr)->joy==e.jaxis.which)&&((*itr)->axis==e.jaxis.axis))
			{
				result|=(*itr)->translate(e,queue);
			}
		}

		//debug::removeMessage();
		return(result);
	}

	bool InputEvaluator::handleMouseButtonEvent(const SDL_Event &e,ControlQueue *queue)
	{
		//debug::addMessage("InputEvaluator::handleMouseButtonEvent()");

		bool result=false;

		for(std::list<MouseButtonTranslator *>::iterator itr=translatorsMouseButton.begin(); itr!=translatorsMouseButton.end(); itr++)
		{
			if((*itr)->button==e.button.button)
			{
				if((((*itr)->state==ctrlUp)&&(e.type==SDL_MOUSEBUTTONUP))||
						(((*itr)->state==ctrlDown)&&(e.type==SDL_MOUSEBUTTONDOWN))||
						(((*itr)->state==ctrlPressed)))
				{
					result|=(*itr)->translate(e,queue);
				}
			}
		}

		//debug::removeMessage();
		return(result);
	}

	bool InputEvaluator::handleMouseAxisEvent(const SDL_Event &e,ControlQueue *queue)
	{
		//debug::addMessage("InputEvaluator::handleMouseAxisEvent()");

		bool result=false;

		for(std::list<MouseAxisTranslator *>::iterator itr=translatorsMouseAxis.begin(); itr!=translatorsMouseAxis.end(); itr++)
		{
			//if((((*itr)->axis==0)&&(e.motion.xrel!=0))||
			//		(((*itr)->axis==1)&&(e.motion.yrel!=0)))
			{
				result|=(*itr)->translate(e,queue);
			}
		}

		//debug::removeMessage();
		return(result);
	}

	std::string InputEvaluator::getNodeName()
	{
		return("inputEvaluator");
	}

	void Trigger::getTranslatorsFromNode(const xml::Node *root)
	{
		//debug::addMessage("Trigger::getTranslatorsFromNode()");

		clearTranslators();

		for(xml::Nodes::const_iterator itr=root->children.begin(); itr!=root->children.end(); itr++)
		{
			const xml::Node *node=*itr;

			if(node->name==XML_KEYTRANSLATOR)
			{
				//debug::addMessage("[Reading key translator]");

				SDLKey key;

				try
				{
					key=strTo<SDLKey>(node->getAttribute(XML_KEY,""));
				}
				catch(std::string ex)
				{
					debug::warning(dbgControls,"%s",ex.c_str());
					continue;
				}

				KeyState state=strTo<KeyState>(node->getAttribute(XML_STATE,"default"));
				GLfloat value=node->getAttributeFloat(XML_VALUE,1);

				if(state==ctrlDefault)
				{
					state=defaultKeyState;
				}

				//debug::addMessage("[Creating translator]");
				new KeyTranslator(this,key,state,value);
				//debug::removeMessage();

				//debug::removeMessage();
			}
			else if(node->name==XML_JOYBUTTONTRANSLATOR)
			{
				//debug::addMessage("[Reading joy button translator]");

				Sint16 joy=node->getAttributeLong(XML_JOY,0);
				Sint16 button=node->getAttributeLong(XML_BUTTON,0);
				KeyState state=strTo<KeyState>(node->getAttribute(XML_STATE,"default"));
				GLfloat value=node->getAttributeFloat(XML_VALUE,1);

				if(state==ctrlDefault)
				{
					state=defaultKeyState;
				}

				//debug::addMessage("[Creating translator]");
				new JoyButtonTranslator(this,joy,button,state,value);
				//debug::removeMessage();

				//debug::removeMessage();
			}
			else if(node->name==XML_JOYAXISTRANSLATOR)
			{
				//debug::addMessage("[Reading joy axis translator]");

				Sint16 joy=node->getAttributeLong(XML_JOY,0);
				Uint8 axis=node->getAttributeLong(XML_AXIS,0);
				short int direction=node->getAttributeLong(XML_DIRECTION,0);
				Sint16 min=node->getAttributeLong(XML_MIN,0);
				Sint16 max=node->getAttributeLong(XML_MAX,1);

				//debug::addMessage("[Creating translator]");
				new JoyAxisTranslator(this,joy,axis,direction,min,max);
				//debug::removeMessage();

				//debug::removeMessage();
			}
			else if(node->name==XML_MOUSEBUTTONTRANSLATOR)
			{
				//debug::addMessage("[Reading mouse button translator]");

				Uint8 button=node->getAttributeLong(XML_BUTTON,0);
				KeyState state=strTo<KeyState>(node->getAttribute(XML_STATE,"default"));
				GLfloat value=node->getAttributeFloat(XML_VALUE,1);

				if(state==ctrlDefault)
				{
					state=defaultKeyState;
				}

				//debug::addMessage("[Creating translator]");
				new MouseButtonTranslator(this,button,state,value);
				//debug::removeMessage();

				//debug::removeMessage();
			}
			else if(node->name==XML_MOUSEAXISTRANSLATOR)
			{
				//debug::addMessage("[Reading mouose axis translator]");

				Uint8 axis=node->getAttributeLong(XML_AXIS,0);
				short int direction=node->getAttributeLong(XML_DIRECTION,0);
				Sint16 min=node->getAttributeLong(XML_MIN,0);
				Sint16 max=node->getAttributeLong(XML_MAX,1);

				//debug::addMessage("[Creating translator]");
				new MouseAxisTranslator(this,axis,direction,min,max);
				//debug::removeMessage();

				//debug::removeMessage();
			}
		}

		//debug::removeMessage();
	}

	void Trigger::getFromNode(const xml::Node *root,const std::string &baseDir)
	{
		//debug::addMessage("Trigger::getFromNode()");

		const std::string *s=root->getAttribute(XML_DEFAULTKEYSTATE);

		if(s)
		{
			defaultKeyState=strTo<KeyState>(*s);
		}

		getTranslatorsFromNode(root);

		//debug::removeMessage();
	}

	xml::Node *Trigger::buildTranslatorTree()
	{
		//debug::addMessage("Trigger::buildTranslatorTree()");

		xml::Node *root=new xml::Node();
		root->name=XML_TRIGGER;
		(*root)[XML_NAME]=name;

		for(std::list<Translator *>::iterator itr=translators.begin(); itr!=translators.end(); itr++)
		{
			(*root)+=(*itr)->buildTranslatorTree();
		}

		//debug::removeMessage();
		return(root);
	}

	void InputEvaluator::getTranslatorsFromNode(const xml::Node *root)
	{
		//debug::addMessage("InputEvaluator::getFromNode()");

		for(xml::Nodes::const_iterator itr=root->children.begin(); itr!=root->children.end(); itr++)
		{
			const xml::Node *node=*itr;

			if((node->name==XML_TRIGGER)||(node->name==XML_AXIS))
			{
				((*this)[node->getAttribute(XML_NAME,"")])->getTranslatorsFromNode(node);
			}
		}

		//debug::removeMessage();
	}

	void InputEvaluator::addTranslatorsToTree(xml::Node *root)
	{
		//debug::addMessage("InputEvaluator::addTranslatorsToTree()");

		for(iterator itr=begin(); itr!=end(); itr++)
		{
			(*root)+=(*itr)->buildTranslatorTree();
		}

		//debug::removeMessage();
	}

	void InputEvaluator::getFromNode(const xml::Node *root,const std::string &baseDir)
	{
		//debug::addMessage("InputEvaluator::getFromNode()");

		for(xml::Nodes::const_iterator itr=root->children.begin(); itr!=root->children.end(); itr++)
		{
			const xml::Node *node=*itr;

			Trigger *newTrigger=0;

			if(node->name==XML_TRIGGER)
			{
				ControlId code=node->getAttributeLong(XML_CODE,0);
				std::string name=node->getAttribute(XML_NAME,"");
				std::string description=node->getAttribute(XML_DESCRIPTION,"");

				newTrigger=new Trigger(this,code,name,description);
				newTrigger->getFromNode(node);

				newTrigger->queue=node->getAttributeBool(XML_QUEUE,newTrigger->queue);
				newTrigger->data=node->getAttributeLong(XML_DATA,newTrigger->data);
				newTrigger->dataString=node->getAttribute(XML_DATASTRING,newTrigger->dataString);
			}
			else if(node->name==XML_AXIS)
			{
				ControlId code=node->getAttributeLong(XML_CODE,0);
				std::string name=node->getAttribute(XML_NAME,"");
				std::string description=node->getAttribute(XML_DESCRIPTION,"");
				GLfloat max=node->getAttributeFloat(XML_MAX,1);

				newTrigger=new Trigger(this,code,name,description,max);
				newTrigger->getFromNode(node);

				newTrigger->queue=node->getAttributeBool(XML_QUEUE,newTrigger->queue);
				newTrigger->data=node->getAttributeLong(XML_DATA,newTrigger->data);
				newTrigger->dataString=node->getAttribute(XML_DATASTRING,newTrigger->dataString);
			}
		}

		//debug::removeMessage();
	}

	std::map<std::string,SDLKey> &getNameKeyMap()
	{
		//debug::addMessage("getNameKeyMap()");

		static std::map<std::string,SDLKey> map=std::map<std::string,SDLKey>();

		if(map.size()==0)
		{
			std::map<SDLKey,std::string> map2=getKeyNameMap();

			for(std::map<SDLKey,std::string>::iterator itr=map2.begin(); itr!=map2.end(); itr++)
			{
				map[itr->second]=itr->first;
			}
		}

		//debug::removeMessage();
		return(map);
	}

	std::map<SDLKey,std::string> &getKeyNameMap()
	{
		//debug::addMessage("getKeyNameMap()");

		static std::map<SDLKey,std::string> map=std::map<SDLKey,std::string>();

		if(map.size()==0)
		{
			map[SDLK_BACKSPACE]="backspace";
			map[SDLK_TAB]="tab";
			map[SDLK_CLEAR]="clear";
			map[SDLK_RETURN]="return";
			map[SDLK_PAUSE]="pause";
			map[SDLK_ESCAPE]="esc";
			map[SDLK_SPACE]="space";
			map[SDLK_EXCLAIM]="!";
			map[SDLK_QUOTEDBL]="\"";
			map[SDLK_HASH]="#";
			map[SDLK_DOLLAR]="$";
			map[SDLK_AMPERSAND]="&";
			map[SDLK_QUOTE]="'";
			map[SDLK_LEFTPAREN]="(";
			map[SDLK_RIGHTPAREN]=")";
			map[SDLK_ASTERISK]="*";
			map[SDLK_PLUS]="+";
			map[SDLK_COMMA]=",";
			map[SDLK_MINUS]="-";
			map[SDLK_PERIOD]=".";
			map[SDLK_SLASH]="/";
			map[SDLK_0]="0";
			map[SDLK_1]="1";
			map[SDLK_2]="2";
			map[SDLK_3]="3";
			map[SDLK_4]="4";
			map[SDLK_5]="5";
			map[SDLK_6]="6";
			map[SDLK_7]="7";
			map[SDLK_8]="8";
			map[SDLK_9]="9";
			map[SDLK_COLON]=":";
			map[SDLK_SEMICOLON]=";";
			map[SDLK_LESS]="<";
			map[SDLK_EQUALS]="=";
			map[SDLK_GREATER]=">";
			map[SDLK_QUESTION]="?";
			map[SDLK_AT]="@";
			map[SDLK_LEFTBRACKET]="[";
			map[SDLK_BACKSLASH]="\\";
			map[SDLK_RIGHTBRACKET]="}";
			map[SDLK_CARET]="^";
			map[SDLK_UNDERSCORE]="_";
			map[SDLK_BACKQUOTE]="`";
			map[SDLK_a]="a";
			map[SDLK_b]="b";
			map[SDLK_c]="c";
			map[SDLK_d]="d";
			map[SDLK_e]="e";
			map[SDLK_f]="f";
			map[SDLK_g]="g";
			map[SDLK_h]="h";
			map[SDLK_i]="i";
			map[SDLK_j]="j";
			map[SDLK_k]="k";
			map[SDLK_l]="l";
			map[SDLK_m]="m";
			map[SDLK_n]="n";
			map[SDLK_o]="o";
			map[SDLK_p]="p";
			map[SDLK_q]="q";
			map[SDLK_r]="r";
			map[SDLK_s]="s";
			map[SDLK_t]="t";
			map[SDLK_u]="u";
			map[SDLK_v]="v";
			map[SDLK_w]="w";
			map[SDLK_x]="x";
			map[SDLK_y]="y";
			map[SDLK_z]="z";
			map[SDLK_DELETE]="del";
			map[SDLK_KP0]="pad_0";
			map[SDLK_KP1]="pad_1";
			map[SDLK_KP2]="pad_2";
			map[SDLK_KP3]="pad_3";
			map[SDLK_KP4]="pad_4";
			map[SDLK_KP5]="pad_5";
			map[SDLK_KP6]="pad_6";
			map[SDLK_KP7]="pad_7";
			map[SDLK_KP8]="pad_8";
			map[SDLK_KP9]="pad_9";
			map[SDLK_KP_PERIOD]="pad_.";
			map[SDLK_KP_DIVIDE]="pad_/";
			map[SDLK_KP_MULTIPLY]="pad_*";
			map[SDLK_KP_MINUS]="pad_-";
			map[SDLK_KP_PLUS]="pad_+";
			map[SDLK_KP_ENTER]="enter";
			map[SDLK_KP_EQUALS]="pad_=";
			map[SDLK_UP]="up";
			map[SDLK_DOWN]="down";
			map[SDLK_RIGHT]="right";
			map[SDLK_LEFT]="left";
			map[SDLK_INSERT]="ins";
			map[SDLK_HOME]="home";
			map[SDLK_END]="end";
			map[SDLK_PAGEUP]="pgup";
			map[SDLK_PAGEDOWN]="pgdn";
			map[SDLK_F1]="f1";
			map[SDLK_F2]="f2";
			map[SDLK_F3]="f3";
			map[SDLK_F4]="f4";
			map[SDLK_F5]="f5";
			map[SDLK_F6]="f6";
			map[SDLK_F7]="f7";
			map[SDLK_F8]="f8";
			map[SDLK_F9]="f9";
			map[SDLK_F10]="f10";
			map[SDLK_F11]="f11";
			map[SDLK_F12]="f12";
			map[SDLK_F13]="f13";
			map[SDLK_F14]="f14";
			map[SDLK_F15]="f15";
			map[SDLK_NUMLOCK]="numlock";
			map[SDLK_CAPSLOCK]="capslock";
			map[SDLK_SCROLLOCK]="scrollock";
			map[SDLK_RSHIFT]="rshift";
			map[SDLK_LSHIFT]="lshift";
			map[SDLK_RCTRL]="rctrl";
			map[SDLK_LCTRL]="lctrl";
			map[SDLK_RALT]="ralt";
			map[SDLK_LALT]="lalt";
			map[SDLK_RMETA]="rmeta";
			map[SDLK_LMETA]="lmeta";
			map[SDLK_LSUPER]="lwin";
			map[SDLK_RSUPER]="rwin";
			map[SDLK_MODE]="mode";
			map[SDLK_HELP]="help";
			map[SDLK_PRINT]="printscreen";
			map[SDLK_SYSREQ]="sysrq";
			map[SDLK_BREAK]="break";
			map[SDLK_MENU]="menu";
			map[SDLK_POWER]="power";
			map[SDLK_EURO]="euro";
		}

		//debug::removeMessage();
		return(map);
	}
};
