/**
 * MySDL - A multimedia framework based on SDL and OpenGL
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://www.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <assert.h>

#include "containers.h"
#include "core.h"
#include "lists.h"
#include "slider.h"
#include "standard.h"
#include "styles.h"
#include "text.h"
#include "xml.h"

namespace mysdl
{
	namespace gui
	{
		// ================================================================
		//  Common functions
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		template<> Orientation strTo<Orientation>(std::string str)
		{
			if(str=="horizontal")
			{
				return(mysdl::gui::orHorizontal);
			}
			else if(str=="vertical")
			{
				return(mysdl::gui::orVertical);
			}
			else
			{
				throw(str);
			}
		}

		// ================================================================
		//  GuiResourceBankGenerator
		// ================================================================

		/**
		 * A resource bank generator for GUIs.
		 */
		class GuiResourceBankGenerator : public ResourceBankGenerator
		{
			public:
				BasicResourceBank *generateBank(const std::string &name) const;
		};

		/**
		 * Generates a GUI if the bank name is "gui".
		 *
		 * @param name Ignored?
		 *
		 * @return A GUI, or 0.
		 */
		BasicResourceBank *GuiResourceBankGenerator::generateBank(const std::string &name) const
		{
			if(name=="gui")
			{
				return(new GUI());
			}
			else
			{
				return(0);
			}
		}

		/** A generator for GUIs (automatically created & registered. */
		GuiResourceBankGenerator resourceBankGenerator;

		// ================================================================
		//  Style
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		void Style::getFromNode(const xml::Node *node,GUI *gui)
		{
			//debug::addMessage("Style::getFromNode()");

			const std::string *name=node->getAttribute(XML_GUI_NAME);

			if(name)
			{
				this->name=*name;
			}
			else
			{
				debug::error(dbgGui,"No name specified for style. Using empty name.");
				this->name="";
			}

			//debug::removeMessage();
		}

		// ================================================================
		//  Component
		// ================================================================

		/** Helper variable used when drawing components. */
		Vector2D currentPos=Vector2D();

		/*
		 * See header file for documentation.
		 */
		Component::Component() : parent(0),children(),focused(false),wantsFocus(false),
			name(),pos(),dimensions(),gridPos(-1,-1),gridDimensions(1,1),visible(true),enabled(true),neighbours(),triggerMove(),
			data(0),dataString()
		{
			neighbours[dirNorth]=neighbours[dirEast]=neighbours[dirSouth]=neighbours[dirWest]=0;
		}

		/*
		 * See header file for documentation.
		 */
		Component::~Component()
		{
			//debug::addMessage("Component::~Component()");

			for(Component *c=children.front(); !children.empty(); c=children.front())
			{
				delete c;
			}

			if(parent)
			{
				(*parent)-=this;
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void Component::bringToFront()
		{
			Component *p=parent;

			if(p)
			{
				p->children.remove(this);
				p->children.push_front(this);
			}
		}

		/*
		 * See header file for documentation.
		 */
		void Component::getComponentFromNode(const xml::Node *node,GUI *gui)
		{
			//debug::addMessage("Component::getComponentFromNode()");

			if(!gui)
			{
				gui=dynamic_cast<GUI *>(this);
			}

			getSelfFromNode(node,gui);
			getChildrenFromNode(node,gui);

			recalcNeighbours();

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void Component::getSelfFromNode(const xml::Node *node,GUI *gui)
		{
			//debug::addMessage("Component::getSelfFromNode()");

			const std::string *name=node->getAttribute(XML_GUI_NAME);

			if(name)
			{
				this->name=*name;
			}
			else
			{
				debug::error(dbgGui,"No name specified for style. Using empty name.");
				this->name="";
			}

			Rect dimensions;

			dimensions.pos.x=node->getAttributeFloat(XML_GUI_X,pos.x);
			dimensions.pos.y=node->getAttributeFloat(XML_GUI_Y,pos.y);

			dimensions.dimensions.x=node->getAttributeFloat(XML_GUI_WIDTH,this->dimensions.x);
			dimensions.dimensions.y=node->getAttributeFloat(XML_GUI_HEIGHT,this->dimensions.y);

			setDimensions(dimensions);

			gridPos.x=node->getAttributeLong(XML_GUI_GRID_X,gridPos.x);
			gridPos.y=node->getAttributeLong(XML_GUI_GRID_Y,gridPos.y);

			gridDimensions.x=node->getAttributeLong(XML_GUI_GRID_WIDTH,this->gridDimensions.x);
			gridDimensions.y=node->getAttributeLong(XML_GUI_GRID_HEIGHT,this->gridDimensions.y);

			visible=node->getAttributeBool(XML_GUI_VISIBLE,visible);
			enabled=node->getAttributeBool(XML_GUI_ENABLED,enabled);

			data=node->getAttributeLong(XML_GUI_DATA,data);
			dataString=node->getAttribute(XML_GUI_DATASTRING,dataString);

			if(const std::string *trigger=node->getAttribute(XML_GUI_TRIGGER_NORTH))
			{
				this->triggerMove[dirNorth]=gui->findTrigger(*trigger);
			}

			if(const std::string *trigger=node->getAttribute(XML_GUI_TRIGGER_EAST))
			{
				this->triggerMove[dirEast]=gui->findTrigger(*trigger);
			}

			if(const std::string *trigger=node->getAttribute(XML_GUI_TRIGGER_SOUTH))
			{
				this->triggerMove[dirSouth]=gui->findTrigger(*trigger);
			}

			if(const std::string *trigger=node->getAttribute(XML_GUI_TRIGGER_WEST))
			{
				this->triggerMove[dirWest]=gui->findTrigger(*trigger);
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void operator++(Direction &d,int)
		{
			switch(d)
			{
				case dirNorth:	d=dirEast;  break;
				case dirEast:		d=dirSouth; break;
				case dirSouth:	d=dirWest;  break;
				case dirWest:		d=dirNorth; break;
				default:				assert(0);
			}
		}

		/*
		 * See header file for documentation.
		 */
		void Component::recalcNeighbours()
		{
			int width=0,height=0;

			for(Components::iterator itr=children.begin(); itr!=children.end(); itr++)
			{
				Component *c=*itr;

				if(c->wantsFocus)
				{
					int wn=c->gridPos.x+c->gridDimensions.x;
					int hn=c->gridPos.y+c->gridDimensions.y;

					if(wn>width)
					{
						width=wn;
					}

					if(hn>height)
					{
						height=hn;
					}
				}
			}

			std::vector<std::vector<Component *> > raster;
			raster.resize(width);

			for(int x=0; x<width; x++)
			{
				raster[x].resize(height);

				for(int y=0; y<height; y++)
				{
					raster[x][y]=0;
				}
			}

			for(Components::iterator itr=children.begin(); itr!=children.end(); itr++)
			{
				Component *c=*itr;

				if(c->wantsFocus)
				{
					for(int x=c->gridPos.x; x<c->gridPos.x+c->gridDimensions.x; x++)
					{
						for(int y=c->gridPos.y; y<c->gridPos.y+c->gridDimensions.y; y++)
						{
							raster[x][y]=c;
						}
					}
				}
			}

			for(Components::iterator itr=children.begin(); itr!=children.end(); itr++)
			{
				Component *c=*itr;

				if(c->wantsFocus)
				{
					int rx=c->gridPos.x;
					int ry=c->gridPos.y;

					if((rx<0)||(ry<0))
					{
						for(Direction d=dirNorth;;d++)
						{
							c->neighbours[d]=0;

							if(d==dirWest)
							{
								break;
							}
						}

						continue;
					}

					c->neighbours[dirNorth]=(ry>0 ? raster[rx][ry-1] : raster[rx][height-1]);
					c->neighbours[dirEast ]=(rx<width-1 ? raster[rx+1][ry] : raster[0][ry]);
					c->neighbours[dirSouth]=(ry<height-1 ? raster[rx][ry+1] : raster[rx][0]);
					c->neighbours[dirWest ]=(rx>0 ? raster[rx-1][ry] : raster[width-1][ry]);

					for(Direction d=dirNorth;;d++)
					{
						if(c->neighbours[d]==c)
						{
							c->neighbours[d]=0;
						}

						if(d==dirWest)
						{
							break;
						}
					}
				}
			}
		}

		/*
		 * See header file for documentation.
		 */
		void Component::operator+=(Component *component)
		{
			//debug::addMessage("Component::operator+=()");

			if(component->parent)
			{
				(*(component->parent))-=component;
			}

			if((component->gridPos.x==-1)&&(component->gridPos.y==-1))
			{
				component->gridPos.x=0;
				component->gridPos.y=size();
			}

			children.push_back(component);
			component->parent=this;

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void Component::operator-=(Component *component)
		{
			//debug::addMessage("Component::operator-=()");

			if(component->parent==this)
			{
				children.remove(component);
				component->parent=0;
			}
			else
			{
				throw std::string("Trying to remove a component which is not a child.");
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		Component *Component::operator[](const std::string name)
		{
			//TODO: Map?
			for(Components::iterator itr=children.begin(); itr!=children.end(); itr++)
			{
				if((*itr)->name==name)
				{
					return((*itr));
				}
			}

			for(Components::iterator itr=children.begin(); itr!=children.end(); itr++)
			{
				if(Component *c=(**itr)[name])
				{
					return(c);
				}
			}

			return(0);
		}

		/*
		 * See header file for documentation.
		 */
		bool Component::joyAxisMotion(ControlQueue *queue,Uint8 joy,Uint8 axis,short int digitalValue)
		{
			if(joy==0)
			{
				if(axis==0)
				{
					if(digitalValue==+1)
					{
						getRoot()->getFocused()->move(dirEast,queue); // why getRoot()?
					}
					else if(digitalValue==-1)
					{
						getRoot()->getFocused()->move(dirWest,queue); // why getRoot()?
					}

					return(true);
				}
				else if(axis==1)
				{
					if(digitalValue==+1)
					{
						getRoot()->getFocused()->move(dirSouth,queue); // why getRoot()?
					}
					else if(digitalValue==-1)
					{
						getRoot()->getFocused()->move(dirNorth,queue); // why getRoot()?
					}

					return(true);
				}
				else
				{
					return(false);
				}
			}

			return(false);
		}

		/*
		 * See header file for documentation.
		 */
		bool Component::keyDown(ControlQueue *queue,SDLKey key,SDLMod modifier,Uint16 unicode)
		{
			if(key==9)
			{
				focusNext();
				return(true);
			}
			else if(key==SDLK_UP)
			{
				getRoot()->getFocused()->move(dirNorth,queue); // why getRoot()?
				return(true);
			}
			else if(key==SDLK_DOWN)
			{
				getRoot()->getFocused()->move(dirSouth,queue); // why getRoot()?
				return(true);
			}
			else if(key==SDLK_LEFT)
			{
				getRoot()->getFocused()->move(dirWest,queue); // why getRoot()?
				return(true);
			}
			else if(key==SDLK_RIGHT)
			{
				getRoot()->getFocused()->move(dirEast,queue); // why getRoot()?
				return(true);
			}
			else
			{
				return(false);
			}
		};

		/*
		 * See header file for documentation.
		 */
		void Component::move(Direction dir,ControlQueue *queue)
		{
			//debug::addMessage("Component::move()");

			if(triggerMove[dir])
			{
				(*queue) << triggerMove[dir];
			}
			else if(neighbours[dir])
			{
				neighbours[dir]->setFocused(true);
				getRoot()->doMove();
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void Component::focusNext()
		{
			if(children.size()>0)
			{
				if(children.front()->wantsFocus)
				{
					children.front()->setFocused(true);
					getRoot()->doMove();
				}
				else
				{
					children.front()->focusNext();
				}
			}
			else if(parent)
			{
				Components::iterator itr=parent->children.begin();

				while((*itr)!=this)
				{
					itr++;
				}

				itr++;

				if(itr!=parent->children.end())
				{
					if((*itr)->wantsFocus)
					{
						(*itr)->setFocused(true);
						getRoot()->doMove();
					}
					else
					{
						(*itr)->focusNext();
					}
				}
				else
				{
					if(parent->wantsFocus)
					{
						parent->setFocused(true);
						getRoot()->doMove();
					}
					else
					{
						parent->focusNext();
					}
				}
			}
		}

		/*
		 * See header file for documentation.
		 */
		void Component::focusPrevious()
		{
		}

		/*
		 * See header file for documentation.
		 */
		void Component::center(const Vector2D dimensions)
		{
			if(dimensions.x>0)
			{
				pos.x=(int)(dimensions.x-this->dimensions.x)/2;
			}

			if(dimensions.y>0)
			{
				pos.y=(int)(dimensions.y-this->dimensions.y)/2;
			}
		}

		/*
		 * See header file for documentation.
		 */
		void Component::draw()
		{
			//debug::addMessage("Component::draw()");

			if(visible)
			{
				currentPos+=pos;
				glTranslatef(pos.x,pos.y,0);

				drawSelf();

				glPushAttrib(GL_SCISSOR_BIT);
				glScissor((GLint)currentPos.x,(GLint)currentPos.y,(GLint)(currentPos.x+dimensions.x),(GLint)(currentPos.y+dimensions.y));
				glEnable(GL_SCISSOR_TEST);

				for(Components::reverse_iterator itr=children.rbegin(); itr!=children.rend(); itr++)
				{
					(*itr)->draw();
				}

				glPopAttrib();

				drawSelfPost();

				glTranslatef(-pos.x,-pos.y,0);
				currentPos-=pos;
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void Component::perform(ControlQueue *queue,GLfloat ticks)
		{
			//debug::addMessage("Component::perform()");

			if(enabled)
			{
				performSelf(queue,ticks);

				for(Components::reverse_iterator itr=children.rbegin(); itr!=children.rend(); itr++)
				{
					(*itr)->perform(queue,ticks);
				}
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		Component *Component::getParent()
		{
			return(parent);
		}

		/*
		 * See header file for documentation.
		 */
		GUI *Component::getRoot()
		{
			//debug::addMessage("Component::getRoot()");

			Component *root=this;

			while(root->parent)
			{
				root=root->parent;
			}

			//debug::removeMessage();

			return(dynamic_cast<GUI *>(root));
		}

		/*
		 * See header file for documentation.
		 */
		void Component::captureMouse(bool capture)
		{
			//debug::addMessage("Component::captureMouse()");

			GUI *root=getRoot();

			if(capture)
			{
				root->mouseCaptured=this;
			}
			else
			{
				if(root->mouseCaptured==this)
				{
					root->mouseCaptured=0;
				}
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void Component::setFocused(bool focused)
		{
			//debug::addMessage("Component::setFocused()");

			GUI *root=getRoot();

			if(((focused)&&(root->focused!=this))||
			   ((!focused)&&(root->focused==this)))
			{
				if(focused)
				{
					if(root->focused)
					{
						root->focused->setFocused(false);
					}
				}

				Component *c=this;

				while(c)
				{
					c->focused=focused;
					c=c->parent;
				}

				root->focused=this;
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		bool Component::isFocused()
		{
			return(focused);
		}

		/*
		 * See header file for documentation.
		 */
		bool Component::contains(Vector2D pos)
		{
			return((pos.x>=0)&&
						 (pos.x<this->dimensions.x)&&
						 (pos.y>=0)&&
						 (pos.y<this->dimensions.y));
		}

		/*
		 * See header file for documentation.
		 */
		bool Component::isEnabled()
		{
			return((enabled)&&(visible));
		}

		/*
		 * See header file for documentation.
		 */
		Component *Component::find(const Vector2D pos)
		{
			if((isEnabled())&&(contains(pos)))
			{
				for(Components::iterator itr=children.begin(); itr!=children.end(); itr++)
				{
					Vector2D subpos=pos-(*itr)->pos;

					if(Component *c=(*itr)->find(subpos))
					{
						return(c);
					}
				}

				return(this);
			}
			else
			{
				return(0);
			}
		}

		/*
		 * See header file for documentation.
		 */
		Vector2D Component::getRelativeCoordinates(Vector2D pos)
		{
			if(parent)
			{
				pos=parent->getRelativeCoordinates(pos);
			}

			return(pos-this->pos);
		}

		/*
		 * See header file for documentation.
		 */
		void Component::getChildrenFromNode(const xml::Node *node,GUI *gui)
		{
			//debug::addMessage("Component::getChildrenFromNode()");

			for(xml::Nodes::const_iterator itr=node->children.begin(); itr!=node->children.end(); itr++)
			{
				xml::Node *node=*itr;

				Component *component=0;

				if(node->name==XML_GUI_PANE)
				{
					component=new Pane();
				}
				else if(node->name==XML_GUI_INPUTCAPTOR)
				{
					component=new InputCaptor();
				}
				else if(node->name==XML_GUI_LABEL)
				{
					component=new Label();
				}
				else if(node->name==XML_GUI_BUTTON)
				{
					component=new Button();
				}
				else if(node->name==XML_GUI_CHECKBOX)
				{
					component=new Checkbox();
				}
				else if(node->name==XML_GUI_DROPDOWNLIST)
				{
					component=new DropdownList();
				}
				else if(node->name==XML_GUI_TEXTFIELD)
				{
					component=new Textfield();
				}
				else if(node->name==XML_GUI_SLIDER_INT)
				{
					component=new Slider<int>();
				}
				else if(node->name==XML_GUI_SLIDER_FLOAT)
				{
					component=new Slider<float>();
				}

				if(component)
				{
					component->getComponentFromNode(node,gui);
					(*this)+=component;
				}
			}

			//debug::removeMessage();
		}

		// ================================================================
		//  GUI
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		GUI::GUI() : Component(),BasicResourceBank(),
			fonts(0),images(0),triggers(0),styles(0),mouseCursor(0),mouseCaptured(0),
			focused(this),active(false),dir(),keydelay(300),keyrepeat(30),callback(0)
		{
			visible=false;
			styles=new StyleBank();
		}

		/*
		 * See header file for documentation.
		 */
		GUI::~GUI()
		{
			if(styles)
			{
				delete styles;
				styles=0;
			}

			if(mouseCursor)
			{
				delete mouseCursor;
				mouseCursor=0;
			}
		}

		/*
		 * See header file for documentation.
		 */
		void GUI::setFonts(fonts::FontBank *fonts)
		{
			this->fonts=fonts;
		}

		/*
		 * See header file for documentation.
		 */
		fonts::FontBank *GUI::getFonts()
		{
			return(fonts);
		}

		/*
		 * See header file for documentation.
		 */
		void GUI::setImages(mysdl::gfx::ImageBank *images)
		{
			this->images=images;
		}

		/*
		 * See header file for documentation.
		 */
		mysdl::gfx::ImageBank *GUI::getImages()
		{
			return(images);
		}

		/*
		 * See header file for documentation.
		 */
		std::map<std::string,Trigger *> *GUI::getTriggers()
		{
			return(triggers);
		}

		/*
		 * See header file for documentation.
		 */
		StyleBank *GUI::getStyles()
		{
			return(styles);
		}

		/*
		 * See header file for documentation.
		 */
		mysdl::sprites::Sprite *GUI::getMouseCursor()
		{
			return(mouseCursor);
		}

		/*
		 * See header file for documentation.
		 */
		Component *GUI::getMouseCaptured()
		{
			return(mouseCaptured);
		}

		/*
		 * See header file for documentation.
		 */
		Component *GUI::getFocused()
		{
			return(focused);
		}

		/*
		 * See header file for documentation.
		 */
		bool GUI::getActive()
		{
			return(active);
		}

		/*
		 * See header file for documentation.
		 */
		GLfloat GUI::getKeyDelay()
		{
			return(keydelay);
		}

		/*
		 * See header file for documentation.
		 */
		GLfloat GUI::getKeyRepeat()
		{
			return(keyrepeat);
		}

		/*
		 * See header file for documentation.
		 */
		void GUI::setMouseCursor(mysdl::sprites::Sprite *mouseCursor)
		{
			this->mouseCursor=mouseCursor;
		}

		/*
		 * See header file for documentation.
		 */
		void GUI::setActive(bool active)
		{
			this->active=active;
		}

		/*
		 * See header file for documentation.
		 */
		void GUI::setKeyRepeat(GLfloat keydelay,GLfloat keyrepeat)
		{
			this->keydelay=keydelay;
			this->keyrepeat=keyrepeat;
		}

		/*
		 * See header file for documentation.
		 */
		Trigger *GUI::findTrigger(const std::string &name) const
		{
			if(!manager)
			{
				debug::critical(dbgGui,"Manager not yet set!");
				throw std::string("Manager not yet set!");
			}

			std::string::size_type i=name.find('.');

			if(i==name.npos)
			{
				debug::critical(dbgGui,"Not a fully qualified trigger name.");
				return(0);
			}

			//debug::addMessage("GUI::findTrigger()");

			std::string evaluatorName=name.substr(0,i);;
			std::string triggerName=name.substr(i+1,name.size()-(i+1));

			if(InputEvaluator *eval=dynamic_cast<InputEvaluator *>((*manager)[evaluatorName]))
			{
				//debug::removeMessage();
				return((*eval)[triggerName]);
			}
			else
			{
				//debug::removeMessage();
				return(0);
			}
		}

		/*
		 * See header file for documentation.
		 */
		void GUI::getSelfFromNode(const xml::Node *node,GUI *gui)
		{
			//debug::addMessage("GUI::getSelfFromNode()");

			if(!gui) // TODO: Allow non-null at all??
			{
				gui=this;
			}

			if(manager)
			{
				if(const std::string *fontBankName=node->getAttribute(XML_GUI_FONTS))
				{
					fonts=(fonts::FontBank *)((*manager)[*fontBankName]);
				}

				if(const std::string *imageBankName=node->getAttribute(XML_GUI_IMAGES))
				{
					images=(gfx::ImageBank *)((*manager)[*imageBankName]);
				}
			}

			Component::getSelfFromNode(node,gui);

			for(xml::Nodes::const_iterator itr=node->children.begin(); itr!=node->children.end(); itr++)
			{
				xml::Node *node=*itr;

				/*if(node->name==XML_GUI_IMAGES)
				{
					images->getFromNode(node,dir);
					images->release();
					images->refresh(0);
				}
				else */if(node->name==XML_GUI_PLAINRECTSTYLE)
				{
					Style *style=new PlainRectStyle();
					style->getFromNode(node,gui);
					(*styles)+=style;
				}
				else if(node->name==XML_GUI_TEXTUREDRECTSTYLE)
				{
					Style *style=new TexturedRectStyle();
					style->getFromNode(node,gui);
					(*styles)+=style;
				}
				else if(node->name==XML_GUI_SIMPLEFONTSTYLE)
				{
					Style *style=new SimpleFontStyle();
					style->getFromNode(node,gui);
					(*styles)+=style;
				}
				else if(node->name==XML_GUI_LABELSTYLE)
				{
					Style *style=new LabelStyle();
					style->getFromNode(node,gui);
					(*styles)+=style;
				}
				else if(node->name==XML_GUI_BUTTONSTYLE)
				{
					Style *style=new ButtonStyle();
					style->getFromNode(node,gui);
					(*styles)+=style;
				}
				else if(node->name==XML_GUI_DROPDOWNLISTSTYLE)
				{
					Style *style=new DropdownListStyle();
					style->getFromNode(node,gui);
					(*styles)+=style;
				}
				else if(node->name==XML_GUI_CHECKBOXSTYLE)
				{
					Style *style=new CheckboxStyle();
					style->getFromNode(node,gui);
					(*styles)+=style;
				}
			}

			if(const std::string *cursor=node->getAttribute(XML_GUI_CURSOR))
			{
				if(mouseCursor)
				{
					delete mouseCursor;
					mouseCursor=0;
				}

				mysdl::sprites::ImageSprite *mouseCursor=new mysdl::sprites::ImageSprite();
				mouseCursor->setImage(dynamic_cast<mysdl::gfx::SingleTextureImage *>((*(images))[*cursor])); // \todo remove cast
				mouseCursor->pos=Vector2D(dimensions.x/2,dimensions.y/2);
				mouseCursor->visible=true;

				this->mouseCursor=mouseCursor;
			}

			//debug::removeMessage();
		}

		/** TODO: ? */
		std::map<std::pair<Uint8,Uint8>,short int> axisTable;

		/*
		 * See header file for documentation.
		 */
		bool GUI::evaluate(const SDL_Event &e,ControlQueue *queue)
		{
			//debug::addMessage("GUI::operator<<(const SDL_Event &)");

			bool result=false;

			if(active)
			{
				switch(e.type)
				{
					case SDL_MOUSEMOTION:
					{
						if(mouseCursor)
						{
							mouseCursor->pos=Vector2D(e.motion.x-pos.x,dimensions.y-e.motion.y-pos.y);

							mouseCursor->pos.x=clamp<GLfloat>(mouseCursor->pos.x,0,dimensions.x-1);
							mouseCursor->pos.y=clamp<GLfloat>(mouseCursor->pos.y,0,dimensions.y-1);

							if(mouseCaptured)
							{
								Component *c=find(mouseCursor->pos);

								if((c==mouseCaptured)&&(focused!=mouseCaptured))
								{
									mouseCaptured->enter(queue,mouseCaptured->getRelativeCoordinates(mouseCursor->pos));
									c->setFocused(true);
								}
								else if((c!=mouseCaptured)&&(focused==mouseCaptured))
								{
									mouseCaptured->leave(queue,mouseCaptured->getRelativeCoordinates(mouseCursor->pos));
									this->setFocused(true);
								}
								else
								{
									mouseCaptured->mouseMove(queue,mouseCaptured->getRelativeCoordinates(mouseCursor->pos));
								}
							}
							else
							{
								Component *newFocused=find(mouseCursor->pos);

								if(newFocused!=focused)
								{
									if(focused)
									{
										focused->leave(queue,mouseCursor->pos);
									}

									if(newFocused)
									{
										newFocused->enter(queue,mouseCursor->pos);
									}

									newFocused->setFocused(true);
								}
								else if(focused)
								{
									focused->mouseMove(queue,focused->getRelativeCoordinates(mouseCursor->pos));
								}
							}
						}

						result=true;

						break;
					}

					case SDL_MOUSEBUTTONDOWN:
					case SDL_MOUSEBUTTONUP:
					{
						Component *c=(mouseCaptured ? mouseCaptured : focused);

						if(c)
						{
							if(e.type==SDL_MOUSEBUTTONDOWN)
							{
								c->mouseDown(queue,c->getRelativeCoordinates(Vector2D(e.button.x,vpScreen.dimensions.y-e.button.y)),e.button.button);
							}
							else
							{
								c->mouseUp(queue,c->getRelativeCoordinates(Vector2D(e.button.x,vpScreen.dimensions.y-e.button.y)),e.button.button);
							}
						}

						result=true;

						break;
					}

					case SDL_KEYUP:
					case SDL_KEYDOWN:
					case SDL_JOYBUTTONUP:
					case SDL_JOYBUTTONDOWN:
					{
						Component *c=focused; // TODO: keyboard capture?

						if(c)
						{
							if(e.type==SDL_KEYDOWN)
							{
								result=c->keyDown(queue,e.key.keysym.sym,e.key.keysym.mod,e.key.keysym.unicode);
							}
							else if(e.type==SDL_KEYUP)
							{
								result=c->keyUp(queue,e.key.keysym.sym,e.key.keysym.mod,e.key.keysym.unicode);
							}
							else if(e.type==SDL_JOYBUTTONDOWN)
							{
								result=c->joyButtonDown(queue,e.jbutton.which,e.jbutton.button);
							}
							else if(e.type==SDL_JOYBUTTONUP)
							{
								result=c->joyButtonUp(queue,e.jbutton.which,e.jbutton.button);
							}
						}

						break;
					}

					case SDL_JOYAXISMOTION:
					{
						Component *c=focused; // TODO: joystick capture?
						short int digitalValue=e.jaxis.value>16000 ? +1 : e.jaxis.value<-16000 ? -1 : 0; // TODO: add border deadzone?

						if(axisTable[std::pair<Uint8,Uint8>(e.jaxis.which,e.jaxis.axis)]!=digitalValue)
						{
							axisTable[std::pair<Uint8,Uint8>(e.jaxis.which,e.jaxis.axis)]=digitalValue;
							result=c->joyAxisMotion(queue,e.jaxis.which,e.jaxis.axis,digitalValue);
						}
						break;
					}
				}
			}

			//debug::removeMessage();

			return(result);
		}

		/*
		 * See header file for documentation.
		 */
		void GUI::draw()
		{
			//debug::addMessage("GUI::draw()");

			glPushAttrib(GL_ALL_ATTRIB_BITS);
			glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
			glDisable(GL_DEPTH_TEST);
			glEnable(GL_BLEND);

			currentPos=Vector2D(0,0);
			Component::draw();

			glPopAttrib();

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void GUI::drawSelfPost()
		{
			if((mouseCursor)&&(mouseCursor->visible))
			{
				mouseCursor->draw();
			}
		}

		/*
		* See header file for documentation.
		*/
		void GUI::setCallback(GUICallback *callback)
		{
			this->callback=callback;
		}

		/*
		 * See header file for documentation.
		 */
		GUICallback *GUI::getCallback()
		{
			return callback;
		}

		/*
		 * See header file for documentation.
		 */
		void GUI::doMove()
		{
			if(callback)
			{
				callback->onMove();
			}
		}

		/*
		 * See header file for documentation.
		 */
		void GUI::doConfirm()
		{
			if(callback)
			{
				callback->onConfirm();
			}
		}

		// ================================================================
		//  CaptionedObject
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		CaptionedObject::CaptionedObject() : caption(),image(0)
		{
		}

		/*
		 * See header file for documentation.
		 */
		CaptionedObject::~CaptionedObject()
		{
			setImage(0); // TODO: why?
		}

		/*
		 * See header file for documentation.
		 */
		void CaptionedObject::setCaption(const std::string &caption)
		{
			this->caption=caption;
		}

		/*
		 * See header file for documentation.
		 */
		void CaptionedObject::setImage(mysdl::gfx::Image *image)
		{
			this->image=image;
		}

		/*
		 * See header file for documentation.
		 */
		std::string CaptionedObject::getCaption()
		{
			return(caption);
		}

		/*
		 * See header file for documentation.
		 */
		mysdl::gfx::Image *CaptionedObject::getImage()
		{
			return(image);
		}

		/*
		 * See header file for documentation.
		 */
		void CaptionedObject::getSelfFromNode(const xml::Node *node,GUI *gui)
		{
			//debug::addMessage("CaptionedObject::getSelfFromNode()");

			caption=node->getAttribute(XML_GUI_CAPTION,caption);

			if(const std::string *str=node->getAttribute(XML_GUI_IMAGE))
			{
				setImage((*gui->getImages())[*str]);
				debug::debug(dbgGui,"Attached image \"%s\": %p (%s)",str->c_str(),image,caption.c_str());
			}

			//debug::removeMessage();
		}
	};
};
