/**
 * MySDL - A multimedia framework based on SDL and OpenGL
 * Copyright (C) 2006 Marian Schedenig (marian@musgit.com)
 * http://www.musgit.com/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "standard.h"
#include "xml.h"

namespace mysdl
{
	namespace gui
	{
		// ================================================================
		//  InputCaptor
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		InputCaptor::InputCaptor() :
			Component(),
			triggerKey(0),triggerJoyButton(0),triggerJoyAxis(0),
			key(),joy(0),joyButton(0),joyAxis(0),joyAxisDirection(0)
		{
		}

		/*
		 * See header file for documentation.
		 */
		void InputCaptor::getSelfFromNode(const xml::Node *node,GUI *gui)
		{
			//debug::addMessage("InputCapturer::getSelfFromNode()");

			Component::getSelfFromNode(node,gui);

			if(const std::string *trigger=node->getAttribute(XML_GUI_TRIGGER_KEY))
			{
				this->triggerKey=gui->findTrigger(*trigger);
			}

			if(const std::string *trigger=node->getAttribute(XML_GUI_TRIGGER_JOYBUTTON))
			{
				this->triggerJoyButton=gui->findTrigger(*trigger);
			}

			if(const std::string *trigger=node->getAttribute(XML_GUI_TRIGGER_JOYAXIS))
			{
				this->triggerJoyAxis=gui->findTrigger(*trigger);
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		bool InputCaptor::keyDown(ControlQueue *queue,SDLKey key,SDLMod modifier,Uint16 unicode)
		{
			if(getEnabled())
			{
				this->key=key;

				if(triggerKey)
				{
					(*queue) << triggerKey;
				}

				setEnabled(false);
			}

			return(true);
		}

		/*
		 * See header file for documentation.
		 */
		bool InputCaptor::joyButtonDown(ControlQueue *queue,Uint8 joy,Uint8 button)
		{
			if(getEnabled())
			{
				this->joy=joy;
				this->joyButton=button;

				if(triggerJoyButton)
				{
					(*queue) << triggerJoyButton;
				}

				setEnabled(false);
			}

			return(true);
		}

		/*
		 * See header file for documentation.
		 */
		bool InputCaptor::joyAxisMotion(ControlQueue *queue,Uint8 joy,Uint8 axis,short int digitalValue)
		{
			if((digitalValue!=0)&&(getEnabled()))
			{
				this->joy=joy;
				this->joyAxis=axis;
				this->joyAxisDirection=digitalValue;

				if(triggerJoyAxis)
				{
					(*queue) << triggerJoyAxis;
				}

				setEnabled(false);
			}

			return(true);
		}

		/*
		 * See header file for documentation.
		 */
		SDLKey InputCaptor::getKey()
		{
			return(key);
		}

		/*
		 * See header file for documentation.
		 */
		Uint8 InputCaptor::getJoy()
		{
			return(joy);
		}

		/*
		 * See header file for documentation.
		 */
		Uint8 InputCaptor::getJoyButton()
		{
			return(joyButton);
		}

		/*
		 * See header file for documentation.
		 */
		Uint8 InputCaptor::getJoyAxis()
		{
			return(joyAxis);
		}

		/*
		 * See header file for documentation.
		 */
		short int InputCaptor::getJoyAxisDirection()
		{
			return(joyAxisDirection);
		}

		// ================================================================
		//  Button
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		Button::Button() : Component(),CaptionedObject(),holding(false),style(0),trigger(0),onClick(0),repeat(false),repeatTimer(0)
		{
			wantsFocus=true;
		}

		/*
		 * See header file for documentation.
		 */
		void Button::getSelfFromNode(const xml::Node *node,GUI *gui)
		{
			//debug::addMessage("Button::getSelfFromNode()");

			Component::getSelfFromNode(node,gui);
			CaptionedObject::getSelfFromNode(node,gui);

			repeat=node->getAttributeBool(XML_GUI_REPEAT,repeat);

			if(const std::string *style=node->getAttribute(XML_GUI_STYLE))
			{
				this->style=dynamic_cast<ButtonStyle *>((*gui->getStyles())[*style]);
			}

			if(const std::string *image=node->getAttribute(XML_GUI_IMAGE))
			{
				//this->image=((*gui->getImages())[*image]);
				setImage((*gui->getImages())[*image]);
			}

			if(const std::string *trigger=node->getAttribute(XML_GUI_TRIGGER))
			{
				this->trigger=gui->findTrigger(*trigger);
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void Button::drawSelf()
		{
			//debug::addMessage("Button::drawSelf()");

			if(style)
			{
				ButtonState state=getState();

				mysdl::Rect rect;
				rect.pos=Vector2D(0,0);
				rect.dimensions=dimensions;
				style->draw(rect,state);

				glPushMatrix();
				style->print(rect,state,caption.c_str(),image);
				glPopMatrix();
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void Button::mouseDown(ControlQueue *queue,Vector2D pos,Uint8 button)
		{
			if(button==1)
			{
				captureMouse(true);
				holding=true;

				if(repeat)
				{
					doClick(queue);
					repeatTimer=getRoot()->getKeyDelay();
				}
			}
		}

		/*
		 * See header file for documentation.
		 */
		void Button::mouseUp(ControlQueue *queue,Vector2D pos,Uint8 button)
		{
			if(button==1)
			{
				if(isDown())
				{
					doClick(queue);
				}

				repeatTimer=0;
				holding=false;
				captureMouse(false);
			}
		}

		/*
		 * See header file for documentation.
		 */
		void Button::performSelf(ControlQueue *queue,GLfloat ticks)
		{
			if(repeatTimer)
			{
				GLfloat keyRepeat=getRoot()->getKeyRepeat();

				repeatTimer-=ticks;

				while(repeatTimer<=0)
				{
					repeatTimer+=keyRepeat;
					doClick(queue);
				}
			}
		}

		/*
		 * See header file for documentation.
		 */
		void Button::doClick(ControlQueue *queue)
		{
			getRoot()->doConfirm();

			if(onClick)
			{
				onClick->handleClickEvent(this);
			}

			if(trigger)
			{
				(*queue)<<trigger;
			}
		}

		/*
		 * See header file for documentation.
		 */
		bool Button::keyDown(ControlQueue *queue,SDLKey key,SDLMod modifier,Uint16 unicode)
		{
			if((unicode==SDLK_RETURN)||(unicode==SDLK_SPACE))
			{
				doClick(queue);
				return(true);
			}
			else
			{
				return(Component::keyDown(queue,key,modifier,unicode));
			}
		}

		/*
		 * See header file for documentation.
		 */
		bool Button::joyButtonDown(ControlQueue *queue,Uint8 joy,Uint8 button)
		{
			if(button==0)
			{
				doClick(queue);
				return(true);
			}
			else
			{
				return(Component::joyButtonDown(queue,joy,button));
			}
		}

		/*
		 * See header file for documentation.
		 */
		ButtonState Button::getState()
		{
			ButtonState state=ButtonState(0);

			if(isDown())
			{
				state=ButtonState(state | bsDown);
			}

			if(focused)
			{
				state=ButtonState(state | bsHot);
			}

			return(state);
		}

		/*
		 * See header file for documentation.
		 */
		bool Button::isDown()
		{
			return((holding)&&(repeat||focused));
		}

		/*
		 * See header file for documentation.
		 */
		void Button::setStyle(ButtonStyle *style)
		{
			this->style=style;
		}

		/*
		 * See header file for documentation.
		 */
		ButtonStyle *Button::getStyle()
		{
			return(style);
		}

		/*
		 * See header file for documentation.
		 */
		void Button::setTrigger(Trigger *trigger)
		{
			this->trigger=trigger;
		}

		/*
		 * See header file for documentation.
		 */
		Trigger *Button::getTrigger()
		{
			return(trigger);
		}

		/*
		 * See header file for documentation.
		 */
		void Button::setOnClick(ClickEventHandler *onClick)
		{
			this->onClick=onClick;
		}

		/*
		 * See header file for documentation.
		 */
		ClickEventHandler *Button::getOnClick()
		{
			return(onClick);
		}

		/*
		 * See header file for documentation.
		 */
		void Button::setRepeat(bool repeat)
		{
			this->repeat=repeat;
			repeatTimer=0;
		}

		/*
		 * See header file for documentation.
		 */
		bool Button::getRepeat()
		{
			return(repeat);
		}

		// ================================================================
		//  Checkbox
		// ================================================================

		/*
		 * See header file for documentation.
		 */
		Checkbox::Checkbox() : Component(),CaptionedObject(),holding(false),checked(false),style(0),triggerOn(0),triggerOff(0)
		{
			wantsFocus=true;
		}

		/*
		 * See header file for documentation.
		 */
		void Checkbox::getSelfFromNode(const xml::Node *node,GUI *gui)
		{
			//debug::addMessage("Checkbox::getSelfFromNode()");

			Component::getSelfFromNode(node,gui);
			CaptionedObject::getSelfFromNode(node,gui);

			checked=node->getAttributeBool(XML_GUI_CHECKED,checked);

			if(const std::string *style=node->getAttribute(XML_GUI_STYLE))
			{
				this->style=dynamic_cast<CheckboxStyle *>((*gui->getStyles())[*style]);
			}

			if(const std::string *image=node->getAttribute(XML_GUI_IMAGE))
			{
				setImage((*gui->getImages())[*image]);
			}

			if(const std::string *trigger=node->getAttribute(XML_GUI_TRIGGER))
			{
				triggerOn=gui->findTrigger(*trigger);
				triggerOff=gui->findTrigger(*trigger);
			}

			if(const std::string *trigger=node->getAttribute(XML_GUI_TRIGGER_ON))
			{
				triggerOn=gui->findTrigger(*trigger);
			}

			if(const std::string *trigger=node->getAttribute(XML_GUI_TRIGGER_OFF))
			{
				triggerOff=gui->findTrigger(*trigger);
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void Checkbox::drawSelf()
		{
			//debug::addMessage("Checkbox::drawSelf()");

			if(style)
			{
				ButtonState state=getState();

				mysdl::Rect rect;
				rect.pos=Vector2D(0,0);
				rect.dimensions=dimensions;

				if(style)
				{
					style->draw(rect,state,checked);

					glPushMatrix();
					style->print(rect,state,checked,caption.c_str(),image);
					glPopMatrix();
				}
				else
				{
					debug::warning(dbgGui,"Checkbox style not assigned.");
				}
			}

			//debug::removeMessage();
		}

		/*
		 * See header file for documentation.
		 */
		void Checkbox::mouseDown(ControlQueue *queue,Vector2D pos,Uint8 button)
		{
			if(button==1)
			{
				captureMouse(true);
				holding=true;
			}
		}

		/*
		 * See header file for documentation.
		 */
		void Checkbox::mouseUp(ControlQueue *queue,Vector2D pos,Uint8 button)
		{
			if(button==1)
			{
				if(isDown())
				{
					doClick(queue);
				}

				holding=false;
				captureMouse(false);
			}
		}

		/*
		 * See header file for documentation.
		 */
		void Checkbox::doClick(ControlQueue *queue)
		{
			getRoot()->doConfirm();

			checked=!checked;

			Trigger *trigger=(checked ? triggerOn : triggerOff);

			if(trigger)
			{
				(*queue)<<trigger;
			}
		}

		/*
		 * See header file for documentation.
		 */
		bool Checkbox::keyDown(ControlQueue *queue,SDLKey key,SDLMod modifier,Uint16 unicode)
		{
			if((key==SDLK_RETURN)||(key==SDLK_SPACE))
			{
				doClick(queue);
				return(true);
			}
			else
			{
				return(Component::keyDown(queue,key,modifier,unicode));
			}
		}

		/*
		 * See header file for documentation.
		 */
		bool Checkbox::joyButtonDown(ControlQueue *queue,Uint8 joy,Uint8 button)
		{
			if(button==0)
			{
				doClick(queue);
				return(true);
			}
			else
			{
				return(Component::joyButtonDown(queue,joy,button));
			}
		}

		/*
		 * See header file for documentation.
		 */
		ButtonState Checkbox::getState()
		{
			ButtonState state=ButtonState(0);

			if(isDown())
			{
				state=ButtonState(state | bsDown);
			}

			if(focused)
			{
				state=ButtonState(state | bsHot);
			}

			return(state);
		}

		/*
		 * See header file for documentation.
		 */
		bool Checkbox::isDown()
		{
			return((holding)&&(focused));
		}

		/*
		 * See header file for documentation.
		 */
		void Checkbox::setChecked(bool checked)
		{
			this->checked=checked;
		}

		/*
		 * See header file for documentation.
		 */
		bool Checkbox::getChecked()
		{
			return(checked);
		}

		/*
		 * See header file for documentation.
		 */
		void Checkbox::setStyle(CheckboxStyle *style)
		{
			this->style=style;
		}

		/*
		 * See header file for documentation.
		 */
		CheckboxStyle *Checkbox::getStyle()
		{
			return(style);
		}

		/*
		 * See header file for documentation.
		 */
		void Checkbox::setTriggerOn(Trigger *triggerOn)
		{
			this->triggerOn=triggerOn;
		}

		/*
		 * See header file for documentation.
		 */
		Trigger *Checkbox::getTriggerOn()
		{
			return(triggerOn);
		}

		/*
		 * See header file for documentation.
		 */
		void Checkbox::setTriggerOff(Trigger *triggerOff)
		{
			this->triggerOff=triggerOff;
		}

		/*
		 * See header file for documentation.
		 */
		Trigger *Checkbox::getTriggerOff()
		{
			return(triggerOff);
		}
	};
};
